# -*- coding: utf-8 -*-
"""
    Content-type helper based on file extension.

    Original code from web2py
    by Massimo Di Pierro <mdipierro@cs.depaul.edu>

    :license: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)
"""

__all__ = ['contenttype']

CONTENT_TYPE = {
    '.load': 'text/html',
    '.123': 'application/vnd.lotus-1-2-3',
    '.3ds': 'image/x-3ds',
    '.3g2': 'video/3gpp',
    '.3ga': 'video/3gpp',
    '.3gp': 'video/3gpp',
    '.3gpp': 'video/3gpp',
    '.602': 'application/x-t602',
    '.669': 'audio/x-mod',
    '.7z': 'application/x-7z-compressed',
    '.a': 'application/x-archive',
    '.aac': 'audio/mp4',
    '.abw': 'application/x-abiword',
    '.abw.crashed': 'application/x-abiword',
    '.abw.gz': 'application/x-abiword',
    '.ac3': 'audio/ac3',
    '.ace': 'application/x-ace',
    '.adb': 'text/x-adasrc',
    '.ads': 'text/x-adasrc',
    '.afm': 'application/x-font-afm',
    '.ag': 'image/x-applix-graphics',
    '.ai': 'application/illustrator',
    '.aif': 'audio/x-aiff',
    '.aifc': 'audio/x-aiff',
    '.aiff': 'audio/x-aiff',
    '.al': 'application/x-perl',
    '.alz': 'application/x-alz',
    '.amr': 'audio/amr',
    '.ani': 'application/x-navi-animation',
    '.anim[1-9j]': 'video/x-anim',
    '.anx': 'application/annodex',
    '.ape': 'audio/x-ape',
    '.arj': 'application/x-arj',
    '.arw': 'image/x-sony-arw',
    '.as': 'application/x-applix-spreadsheet',
    '.asc': 'text/plain',
    '.asf': 'video/x-ms-asf',
    '.asp': 'application/x-asp',
    '.ass': 'text/x-ssa',
    '.asx': 'audio/x-ms-asx',
    '.atom': 'application/atom+xml',
    '.au': 'audio/basic',
    '.avi': 'video/x-msvideo',
    '.aw': 'application/x-applix-word',
    '.awb': 'audio/amr-wb',
    '.awk': 'application/x-awk',
    '.axa': 'audio/annodex',
    '.axv': 'video/annodex',
    '.bak': 'application/x-trash',
    '.bcpio': 'application/x-bcpio',
    '.bdf': 'application/x-font-bdf',
    '.bib': 'text/x-bibtex',
    '.bin': 'application/octet-stream',
    '.blend': 'application/x-blender',
    '.blender': 'application/x-blender',
    '.bmp': 'image/bmp',
    '.bz': 'application/x-bzip',
    '.bz2': 'application/x-bzip',
    '.c': 'text/x-csrc',
    '.c++': 'text/x-c++src',
    '.cab': 'application/vnd.ms-cab-compressed',
    '.cb7': 'application/x-cb7',
    '.cbr': 'application/x-cbr',
    '.cbt': 'application/x-cbt',
    '.cbz': 'application/x-cbz',
    '.cc': 'text/x-c++src',
    '.cdf': 'application/x-netcdf',
    '.cdr': 'application/vnd.corel-draw',
    '.cer': 'application/x-x509-ca-cert',
    '.cert': 'application/x-x509-ca-cert',
    '.cgm': 'image/cgm',
    '.chm': 'application/x-chm',
    '.chrt': 'application/x-kchart',
    '.class': 'application/x-java',
    '.cls': 'text/x-tex',
    '.cmake': 'text/x-cmake',
    '.cpio': 'application/x-cpio',
    '.cpio.gz': 'application/x-cpio-compressed',
    '.cpp': 'text/x-c++src',
    '.cr2': 'image/x-canon-cr2',
    '.crt': 'application/x-x509-ca-cert',
    '.crw': 'image/x-canon-crw',
    '.cs': 'text/x-csharp',
    '.csh': 'application/x-csh',
    '.css': 'text/css',
    '.cssl': 'text/css',
    '.csv': 'text/csv',
    '.cue': 'application/x-cue',
    '.cur': 'image/x-win-bitmap',
    '.cxx': 'text/x-c++src',
    '.d': 'text/x-dsrc',
    '.dar': 'application/x-dar',
    '.dbf': 'application/x-dbf',
    '.dc': 'application/x-dc-rom',
    '.dcl': 'text/x-dcl',
    '.dcm': 'application/dicom',
    '.dcr': 'image/x-kodak-dcr',
    '.dds': 'image/x-dds',
    '.deb': 'application/x-deb',
    '.der': 'application/x-x509-ca-cert',
    '.desktop': 'application/x-desktop',
    '.dia': 'application/x-dia-diagram',
    '.diff': 'text/x-patch',
    '.divx': 'video/x-msvideo',
    '.djv': 'image/vnd.djvu',
    '.djvu': 'image/vnd.djvu',
    '.dng': 'image/x-adobe-dng',
    '.doc': 'application/msword',
    '.docbook': 'application/docbook+xml',
    '.docm': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
    '.docx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
    '.dot': 'text/vnd.graphviz',
    '.dsl': 'text/x-dsl',
    '.dtd': 'application/xml-dtd',
    '.dtx': 'text/x-tex',
    '.dv': 'video/dv',
    '.dvi': 'application/x-dvi',
    '.dvi.bz2': 'application/x-bzdvi',
    '.dvi.gz': 'application/x-gzdvi',
    '.dwg': 'image/vnd.dwg',
    '.dxf': 'image/vnd.dxf',
    '.e': 'text/x-eiffel',
    '.egon': 'application/x-egon',
    '.eif': 'text/x-eiffel',
    '.el': 'text/x-emacs-lisp',
    '.emf': 'image/x-emf',
    '.emp': 'application/vnd.emusic-emusic_package',
    '.ent': 'application/xml-external-parsed-entity',
    '.eps': 'image/x-eps',
    '.eps.bz2': 'image/x-bzeps',
    '.eps.gz': 'image/x-gzeps',
    '.epsf': 'image/x-eps',
    '.epsf.bz2': 'image/x-bzeps',
    '.epsf.gz': 'image/x-gzeps',
    '.epsi': 'image/x-eps',
    '.epsi.bz2': 'image/x-bzeps',
    '.epsi.gz': 'image/x-gzeps',
    '.epub': 'application/epub+zip',
    '.erl': 'text/x-erlang',
    '.es': 'application/ecmascript',
    '.etheme': 'application/x-e-theme',
    '.etx': 'text/x-setext',
    '.exe': 'application/x-ms-dos-executable',
    '.exr': 'image/x-exr',
    '.ez': 'application/andrew-inset',
    '.f': 'text/x-fortran',
    '.f90': 'text/x-fortran',
    '.f95': 'text/x-fortran',
    '.fb2': 'application/x-fictionbook+xml',
    '.fig': 'image/x-xfig',
    '.fits': 'image/fits',
    '.fl': 'application/x-fluid',
    '.flac': 'audio/x-flac',
    '.flc': 'video/x-flic',
    '.fli': 'video/x-flic',
    '.flv': 'video/x-flv',
    '.flw': 'application/x-kivio',
    '.fo': 'text/x-xslfo',
    '.for': 'text/x-fortran',
    '.g3': 'image/fax-g3',
    '.gb': 'application/x-gameboy-rom',
    '.gba': 'application/x-gba-rom',
    '.gcrd': 'text/directory',
    '.ged': 'application/x-gedcom',
    '.gedcom': 'application/x-gedcom',
    '.gen': 'application/x-genesis-rom',
    '.gf': 'application/x-tex-gf',
    '.gg': 'application/x-sms-rom',
    '.gif': 'image/gif',
    '.glade': 'application/x-glade',
    '.gmo': 'application/x-gettext-translation',
    '.gnc': 'application/x-gnucash',
    '.gnd': 'application/gnunet-directory',
    '.gnucash': 'application/x-gnucash',
    '.gnumeric': 'application/x-gnumeric',
    '.gnuplot': 'application/x-gnuplot',
    '.gp': 'application/x-gnuplot',
    '.gpg': 'application/pgp-encrypted',
    '.gplt': 'application/x-gnuplot',
    '.gra': 'application/x-graphite',
    '.gsf': 'application/x-font-type1',
    '.gsm': 'audio/x-gsm',
    '.gtar': 'application/x-tar',
    '.gv': 'text/vnd.graphviz',
    '.gvp': 'text/x-google-video-pointer',
    '.gz': 'application/x-gzip',
    '.h': 'text/x-chdr',
    '.h++': 'text/x-c++hdr',
    '.hdf': 'application/x-hdf',
    '.hh': 'text/x-c++hdr',
    '.hp': 'text/x-c++hdr',
    '.hpgl': 'application/vnd.hp-hpgl',
    '.hpp': 'text/x-c++hdr',
    '.hs': 'text/x-haskell',
    '.htm': 'text/html',
    '.html': 'text/html',
    '.hwp': 'application/x-hwp',
    '.hwt': 'application/x-hwt',
    '.hxx': 'text/x-c++hdr',
    '.ica': 'application/x-ica',
    '.icb': 'image/x-tga',
    '.icns': 'image/x-icns',
    '.ico': 'image/vnd.microsoft.icon',
    '.ics': 'text/calendar',
    '.idl': 'text/x-idl',
    '.ief': 'image/ief',
    '.iff': 'image/x-iff',
    '.ilbm': 'image/x-ilbm',
    '.ime': 'text/x-imelody',
    '.imy': 'text/x-imelody',
    '.ins': 'text/x-tex',
    '.iptables': 'text/x-iptables',
    '.iso': 'application/x-cd-image',
    '.iso9660': 'application/x-cd-image',
    '.it': 'audio/x-it',
    '.j2k': 'image/jp2',
    '.jad': 'text/vnd.sun.j2me.app-descriptor',
    '.jar': 'application/x-java-archive',
    '.java': 'text/x-java',
    '.jng': 'image/x-jng',
    '.jnlp': 'application/x-java-jnlp-file',
    '.jp2': 'image/jp2',
    '.jpc': 'image/jp2',
    '.jpe': 'image/jpeg',
    '.jpeg': 'image/jpeg',
    '.jpf': 'image/jp2',
    '.jpg': 'image/jpeg',
    '.jpr': 'application/x-jbuilder-project',
    '.jpx': 'image/jp2',
    '.js': 'application/javascript',
    '.json': 'application/json',
    '.jsonp': 'application/jsonp',
    '.k25': 'image/x-kodak-k25',
    '.kar': 'audio/midi',
    '.karbon': 'application/x-karbon',
    '.kdc': 'image/x-kodak-kdc',
    '.kdelnk': 'application/x-desktop',
    '.kexi': 'application/x-kexiproject-sqlite3',
    '.kexic': 'application/x-kexi-connectiondata',
    '.kexis': 'application/x-kexiproject-shortcut',
    '.kfo': 'application/x-kformula',
    '.kil': 'application/x-killustrator',
    '.kino': 'application/smil',
    '.kml': 'application/vnd.google-earth.kml+xml',
    '.kmz': 'application/vnd.google-earth.kmz',
    '.kon': 'application/x-kontour',
    '.kpm': 'application/x-kpovmodeler',
    '.kpr': 'application/x-kpresenter',
    '.kpt': 'application/x-kpresenter',
    '.kra': 'application/x-krita',
    '.ksp': 'application/x-kspread',
    '.kud': 'application/x-kugar',
    '.kwd': 'application/x-kword',
    '.kwt': 'application/x-kword',
    '.la': 'application/x-shared-library-la',
    '.latex': 'text/x-tex',
    '.ldif': 'text/x-ldif',
    '.lha': 'application/x-lha',
    '.lhs': 'text/x-literate-haskell',
    '.lhz': 'application/x-lhz',
    '.log': 'text/x-log',
    '.ltx': 'text/x-tex',
    '.lua': 'text/x-lua',
    '.lwo': 'image/x-lwo',
    '.lwob': 'image/x-lwo',
    '.lws': 'image/x-lws',
    '.ly': 'text/x-lilypond',
    '.lyx': 'application/x-lyx',
    '.lz': 'application/x-lzip',
    '.lzh': 'application/x-lha',
    '.lzma': 'application/x-lzma',
    '.lzo': 'application/x-lzop',
    '.m': 'text/x-matlab',
    '.m15': 'audio/x-mod',
    '.m2t': 'video/mpeg',
    '.m3u': 'audio/x-mpegurl',
    '.m3u8': 'audio/x-mpegurl',
    '.m4': 'application/x-m4',
    '.m4a': 'audio/mp4',
    '.m4b': 'audio/x-m4b',
    '.m4v': 'video/mp4',
    '.mab': 'application/x-markaby',
    '.man': 'application/x-troff-man',
    '.mbox': 'application/mbox',
    '.md': 'application/x-genesis-rom',
    '.mdb': 'application/vnd.ms-access',
    '.mdi': 'image/vnd.ms-modi',
    '.me': 'text/x-troff-me',
    '.med': 'audio/x-mod',
    '.metalink': 'application/metalink+xml',
    '.mgp': 'application/x-magicpoint',
    '.mid': 'audio/midi',
    '.midi': 'audio/midi',
    '.mif': 'application/x-mif',
    '.minipsf': 'audio/x-minipsf',
    '.mka': 'audio/x-matroska',
    '.mkv': 'video/x-matroska',
    '.ml': 'text/x-ocaml',
    '.mli': 'text/x-ocaml',
    '.mm': 'text/x-troff-mm',
    '.mmf': 'application/x-smaf',
    '.mml': 'text/mathml',
    '.mng': 'video/x-mng',
    '.mo': 'application/x-gettext-translation',
    '.mo3': 'audio/x-mo3',
    '.moc': 'text/x-moc',
    '.mod': 'audio/x-mod',
    '.mof': 'text/x-mof',
    '.moov': 'video/quicktime',
    '.mov': 'video/quicktime',
    '.movie': 'video/x-sgi-movie',
    '.mp+': 'audio/x-musepack',
    '.mp2': 'video/mpeg',
    '.mp3': 'audio/mpeg',
    '.mp4': 'video/mp4',
    '.mpc': 'audio/x-musepack',
    '.mpe': 'video/mpeg',
    '.mpeg': 'video/mpeg',
    '.mpg': 'video/mpeg',
    '.mpga': 'audio/mpeg',
    '.mpp': 'audio/x-musepack',
    '.mrl': 'text/x-mrml',
    '.mrml': 'text/x-mrml',
    '.mrw': 'image/x-minolta-mrw',
    '.ms': 'text/x-troff-ms',
    '.msi': 'application/x-msi',
    '.msod': 'image/x-msod',
    '.msx': 'application/x-msx-rom',
    '.mtm': 'audio/x-mod',
    '.mup': 'text/x-mup',
    '.mxf': 'application/mxf',
    '.n64': 'application/x-n64-rom',
    '.nb': 'application/mathematica',
    '.nc': 'application/x-netcdf',
    '.nds': 'application/x-nintendo-ds-rom',
    '.nef': 'image/x-nikon-nef',
    '.nes': 'application/x-nes-rom',
    '.nfo': 'text/x-nfo',
    '.not': 'text/x-mup',
    '.nsc': 'application/x-netshow-channel',
    '.nsv': 'video/x-nsv',
    '.o': 'application/x-object',
    '.obj': 'application/x-tgif',
    '.ocl': 'text/x-ocl',
    '.oda': 'application/oda',
    '.odb': 'application/vnd.oasis.opendocument.database',
    '.odc': 'application/vnd.oasis.opendocument.chart',
    '.odf': 'application/vnd.oasis.opendocument.formula',
    '.odg': 'application/vnd.oasis.opendocument.graphics',
    '.odi': 'application/vnd.oasis.opendocument.image',
    '.odm': 'application/vnd.oasis.opendocument.text-master',
    '.odp': 'application/vnd.oasis.opendocument.presentation',
    '.ods': 'application/vnd.oasis.opendocument.spreadsheet',
    '.odt': 'application/vnd.oasis.opendocument.text',
    '.oga': 'audio/ogg',
    '.ogg': 'video/x-theora+ogg',
    '.ogm': 'video/x-ogm+ogg',
    '.ogv': 'video/ogg',
    '.ogx': 'application/ogg',
    '.old': 'application/x-trash',
    '.oleo': 'application/x-oleo',
    '.opml': 'text/x-opml+xml',
    '.ora': 'image/openraster',
    '.orf': 'image/x-olympus-orf',
    '.otc': 'application/vnd.oasis.opendocument.chart-template',
    '.otf': 'application/x-font-otf',
    '.otg': 'application/vnd.oasis.opendocument.graphics-template',
    '.oth': 'application/vnd.oasis.opendocument.text-web',
    '.otp': 'application/vnd.oasis.opendocument.presentation-template',
    '.ots': 'application/vnd.oasis.opendocument.spreadsheet-template',
    '.ott': 'application/vnd.oasis.opendocument.text-template',
    '.owl': 'application/rdf+xml',
    '.oxt': 'application/vnd.openofficeorg.extension',
    '.p': 'text/x-pascal',
    '.p10': 'application/pkcs10',
    '.p12': 'application/x-pkcs12',
    '.p7b': 'application/x-pkcs7-certificates',
    '.p7s': 'application/pkcs7-signature',
    '.pack': 'application/x-java-pack200',
    '.pak': 'application/x-pak',
    '.par2': 'application/x-par2',
    '.pas': 'text/x-pascal',
    '.patch': 'text/x-patch',
    '.pbm': 'image/x-portable-bitmap',
    '.pcd': 'image/x-photo-cd',
    '.pcf': 'application/x-cisco-vpn-settings',
    '.pcf.gz': 'application/x-font-pcf',
    '.pcf.z': 'application/x-font-pcf',
    '.pcl': 'application/vnd.hp-pcl',
    '.pcx': 'image/x-pcx',
    '.pdb': 'chemical/x-pdb',
    '.pdc': 'application/x-aportisdoc',
    '.pdf': 'application/pdf',
    '.pdf.bz2': 'application/x-bzpdf',
    '.pdf.gz': 'application/x-gzpdf',
    '.pef': 'image/x-pentax-pef',
    '.pem': 'application/x-x509-ca-cert',
    '.perl': 'application/x-perl',
    '.pfa': 'application/x-font-type1',
    '.pfb': 'application/x-font-type1',
    '.pfx': 'application/x-pkcs12',
    '.pgm': 'image/x-portable-graymap',
    '.pgn': 'application/x-chess-pgn',
    '.pgp': 'application/pgp-encrypted',
    '.php': 'application/x-php',
    '.php3': 'application/x-php',
    '.php4': 'application/x-php',
    '.pict': 'image/x-pict',
    '.pict1': 'image/x-pict',
    '.pict2': 'image/x-pict',
    '.pickle': 'application/python-pickle',
    '.pk': 'application/x-tex-pk',
    '.pkipath': 'application/pkix-pkipath',
    '.pkr': 'application/pgp-keys',
    '.pl': 'application/x-perl',
    '.pla': 'audio/x-iriver-pla',
    '.pln': 'application/x-planperfect',
    '.pls': 'audio/x-scpls',
    '.pm': 'application/x-perl',
    '.png': 'image/png',
    '.pnm': 'image/x-portable-anymap',
    '.pntg': 'image/x-macpaint',
    '.po': 'text/x-gettext-translation',
    '.por': 'application/x-spss-por',
    '.pot': 'text/x-gettext-translation-template',
    '.ppm': 'image/x-portable-pixmap',
    '.pps': 'application/vnd.ms-powerpoint',
    '.ppt': 'application/vnd.ms-powerpoint',
    '.pptm': 'application/vnd.openxmlformats-officedocument.presentationml.presentation',
    '.pptx': 'application/vnd.openxmlformats-officedocument.presentationml.presentation',
    '.ppz': 'application/vnd.ms-powerpoint',
    '.prc': 'application/x-palm-database',
    '.ps': 'application/postscript',
    '.ps.bz2': 'application/x-bzpostscript',
    '.ps.gz': 'application/x-gzpostscript',
    '.psd': 'image/vnd.adobe.photoshop',
    '.psf': 'audio/x-psf',
    '.psf.gz': 'application/x-gz-font-linux-psf',
    '.psflib': 'audio/x-psflib',
    '.psid': 'audio/prs.sid',
    '.psw': 'application/x-pocket-word',
    '.pw': 'application/x-pw',
    '.py': 'text/x-python',
    '.pyc': 'application/x-python-bytecode',
    '.pyo': 'application/x-python-bytecode',
    '.qif': 'image/x-quicktime',
    '.qt': 'video/quicktime',
    '.qtif': 'image/x-quicktime',
    '.qtl': 'application/x-quicktime-media-link',
    '.qtvr': 'video/quicktime',
    '.ra': 'audio/vnd.rn-realaudio',
    '.raf': 'image/x-fuji-raf',
    '.ram': 'application/ram',
    '.rar': 'application/x-rar',
    '.ras': 'image/x-cmu-raster',
    '.raw': 'image/x-panasonic-raw',
    '.rax': 'audio/vnd.rn-realaudio',
    '.rb': 'application/x-ruby',
    '.rdf': 'application/rdf+xml',
    '.rdfs': 'application/rdf+xml',
    '.reg': 'text/x-ms-regedit',
    '.rej': 'application/x-reject',
    '.rgb': 'image/x-rgb',
    '.rle': 'image/rle',
    '.rm': 'application/vnd.rn-realmedia',
    '.rmj': 'application/vnd.rn-realmedia',
    '.rmm': 'application/vnd.rn-realmedia',
    '.rms': 'application/vnd.rn-realmedia',
    '.rmvb': 'application/vnd.rn-realmedia',
    '.rmx': 'application/vnd.rn-realmedia',
    '.roff': 'text/troff',
    '.rp': 'image/vnd.rn-realpix',
    '.rpm': 'application/x-rpm',
    '.rss': 'application/rss+xml',
    '.rt': 'text/vnd.rn-realtext',
    '.rtf': 'application/rtf',
    '.rtx': 'text/richtext',
    '.rv': 'video/vnd.rn-realvideo',
    '.rvx': 'video/vnd.rn-realvideo',
    '.s3m': 'audio/x-s3m',
    '.sam': 'application/x-amipro',
    '.sami': 'application/x-sami',
    '.sav': 'application/x-spss-sav',
    '.scm': 'text/x-scheme',
    '.sda': 'application/vnd.stardivision.draw',
    '.sdc': 'application/vnd.stardivision.calc',
    '.sdd': 'application/vnd.stardivision.impress',
    '.sdp': 'application/sdp',
    '.sds': 'application/vnd.stardivision.chart',
    '.sdw': 'application/vnd.stardivision.writer',
    '.sgf': 'application/x-go-sgf',
    '.sgi': 'image/x-sgi',
    '.sgl': 'application/vnd.stardivision.writer',
    '.sgm': 'text/sgml',
    '.sgml': 'text/sgml',
    '.sh': 'application/x-shellscript',
    '.shar': 'application/x-shar',
    '.shn': 'application/x-shorten',
    '.siag': 'application/x-siag',
    '.sid': 'audio/prs.sid',
    '.sik': 'application/x-trash',
    '.sis': 'application/vnd.symbian.install',
    '.sisx': 'x-epoc/x-sisx-app',
    '.sit': 'application/x-stuffit',
    '.siv': 'application/sieve',
    '.sk': 'image/x-skencil',
    '.sk1': 'image/x-skencil',
    '.skr': 'application/pgp-keys',
    '.slk': 'text/spreadsheet',
    '.smaf': 'application/x-smaf',
    '.smc': 'application/x-snes-rom',
    '.smd': 'application/vnd.stardivision.mail',
    '.smf': 'application/vnd.stardivision.math',
    '.smi': 'application/x-sami',
    '.smil': 'application/smil',
    '.sml': 'application/smil',
    '.sms': 'application/x-sms-rom',
    '.snd': 'audio/basic',
    '.so': 'application/x-sharedlib',
    '.spc': 'application/x-pkcs7-certificates',
    '.spd': 'application/x-font-speedo',
    '.spec': 'text/x-rpm-spec',
    '.spl': 'application/x-shockwave-flash',
    '.spx': 'audio/x-speex',
    '.sql': 'text/x-sql',
    '.sr2': 'image/x-sony-sr2',
    '.src': 'application/x-wais-source',
    '.srf': 'image/x-sony-srf',
    '.srt': 'application/x-subrip',
    '.ssa': 'text/x-ssa',
    '.stc': 'application/vnd.sun.xml.calc.template',
    '.std': 'application/vnd.sun.xml.draw.template',
    '.sti': 'application/vnd.sun.xml.impress.template',
    '.stm': 'audio/x-stm',
    '.stw': 'application/vnd.sun.xml.writer.template',
    '.sty': 'text/x-tex',
    '.sub': 'text/x-subviewer',
    '.sun': 'image/x-sun-raster',
    '.sv4cpio': 'application/x-sv4cpio',
    '.sv4crc': 'application/x-sv4crc',
    '.svg': 'image/svg+xml',
    '.svgz': 'image/svg+xml-compressed',
    '.swf': 'application/x-shockwave-flash',
    '.sxc': 'application/vnd.sun.xml.calc',
    '.sxd': 'application/vnd.sun.xml.draw',
    '.sxg': 'application/vnd.sun.xml.writer.global',
    '.sxi': 'application/vnd.sun.xml.impress',
    '.sxm': 'application/vnd.sun.xml.math',
    '.sxw': 'application/vnd.sun.xml.writer',
    '.sylk': 'text/spreadsheet',
    '.t': 'text/troff',
    '.t2t': 'text/x-txt2tags',
    '.tar': 'application/x-tar',
    '.tar.bz': 'application/x-bzip-compressed-tar',
    '.tar.bz2': 'application/x-bzip-compressed-tar',
    '.tar.gz': 'application/x-compressed-tar',
    '.tar.lzma': 'application/x-lzma-compressed-tar',
    '.tar.lzo': 'application/x-tzo',
    '.tar.xz': 'application/x-xz-compressed-tar',
    '.tar.z': 'application/x-tarz',
    '.tbz': 'application/x-bzip-compressed-tar',
    '.tbz2': 'application/x-bzip-compressed-tar',
    '.tcl': 'text/x-tcl',
    '.tex': 'text/x-tex',
    '.texi': 'text/x-texinfo',
    '.texinfo': 'text/x-texinfo',
    '.tga': 'image/x-tga',
    '.tgz': 'application/x-compressed-tar',
    '.theme': 'application/x-theme',
    '.themepack': 'application/x-windows-themepack',
    '.tif': 'image/tiff',
    '.tiff': 'image/tiff',
    '.tk': 'text/x-tcl',
    '.tlz': 'application/x-lzma-compressed-tar',
    '.tnef': 'application/vnd.ms-tnef',
    '.tnf': 'application/vnd.ms-tnef',
    '.toc': 'application/x-cdrdao-toc',
    '.torrent': 'application/x-bittorrent',
    '.tpic': 'image/x-tga',
    '.tr': 'text/troff',
    '.ts': 'application/x-linguist',
    '.tsv': 'text/tab-separated-values',
    '.tta': 'audio/x-tta',
    '.ttc': 'application/x-font-ttf',
    '.ttf': 'application/x-font-ttf',
    '.ttx': 'application/x-font-ttx',
    '.txt': 'text/plain',
    '.txz': 'application/x-xz-compressed-tar',
    '.tzo': 'application/x-tzo',
    '.ufraw': 'application/x-ufraw',
    '.ui': 'application/x-designer',
    '.uil': 'text/x-uil',
    '.ult': 'audio/x-mod',
    '.uni': 'audio/x-mod',
    '.uri': 'text/x-uri',
    '.url': 'text/x-uri',
    '.ustar': 'application/x-ustar',
    '.vala': 'text/x-vala',
    '.vapi': 'text/x-vala',
    '.vcf': 'text/directory',
    '.vcs': 'text/calendar',
    '.vct': 'text/directory',
    '.vda': 'image/x-tga',
    '.vhd': 'text/x-vhdl',
    '.vhdl': 'text/x-vhdl',
    '.viv': 'video/vivo',
    '.vivo': 'video/vivo',
    '.vlc': 'audio/x-mpegurl',
    '.vob': 'video/mpeg',
    '.voc': 'audio/x-voc',
    '.vor': 'application/vnd.stardivision.writer',
    '.vst': 'image/x-tga',
    '.wav': 'audio/x-wav',
    '.wax': 'audio/x-ms-asx',
    '.wb1': 'application/x-quattropro',
    '.wb2': 'application/x-quattropro',
    '.wb3': 'application/x-quattropro',
    '.wbmp': 'image/vnd.wap.wbmp',
    '.wcm': 'application/vnd.ms-works',
    '.wdb': 'application/vnd.ms-works',
    '.webm': 'video/webm',
    '.wk1': 'application/vnd.lotus-1-2-3',
    '.wk3': 'application/vnd.lotus-1-2-3',
    '.wk4': 'application/vnd.lotus-1-2-3',
    '.wks': 'application/vnd.ms-works',
    '.wma': 'audio/x-ms-wma',
    '.wmf': 'image/x-wmf',
    '.wml': 'text/vnd.wap.wml',
    '.wmls': 'text/vnd.wap.wmlscript',
    '.wmv': 'video/x-ms-wmv',
    '.wmx': 'audio/x-ms-asx',
    '.wp': 'application/vnd.wordperfect',
    '.wp4': 'application/vnd.wordperfect',
    '.wp5': 'application/vnd.wordperfect',
    '.wp6': 'application/vnd.wordperfect',
    '.wpd': 'application/vnd.wordperfect',
    '.wpg': 'application/x-wpg',
    '.wpl': 'application/vnd.ms-wpl',
    '.wpp': 'application/vnd.wordperfect',
    '.wps': 'application/vnd.ms-works',
    '.wri': 'application/x-mswrite',
    '.wrl': 'model/vrml',
    '.wv': 'audio/x-wavpack',
    '.wvc': 'audio/x-wavpack-correction',
    '.wvp': 'audio/x-wavpack',
    '.wvx': 'audio/x-ms-asx',
    '.x3f': 'image/x-sigma-x3f',
    '.xac': 'application/x-gnucash',
    '.xbel': 'application/x-xbel',
    '.xbl': 'application/xml',
    '.xbm': 'image/x-xbitmap',
    '.xcf': 'image/x-xcf',
    '.xcf.bz2': 'image/x-compressed-xcf',
    '.xcf.gz': 'image/x-compressed-xcf',
    '.xhtml': 'application/xhtml+xml',
    '.xi': 'audio/x-xi',
    '.xla': 'application/vnd.ms-excel',
    '.xlc': 'application/vnd.ms-excel',
    '.xld': 'application/vnd.ms-excel',
    '.xlf': 'application/x-xliff',
    '.xliff': 'application/x-xliff',
    '.xll': 'application/vnd.ms-excel',
    '.xlm': 'application/vnd.ms-excel',
    '.xls': 'application/vnd.ms-excel',
    '.xlsm': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
    '.xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
    '.xlt': 'application/vnd.ms-excel',
    '.xlw': 'application/vnd.ms-excel',
    '.xm': 'audio/x-xm',
    '.xmf': 'audio/x-xmf',
    '.xmi': 'text/x-xmi',
    '.xml': 'application/xml',
    '.xpm': 'image/x-xpixmap',
    '.xps': 'application/vnd.ms-xpsdocument',
    '.xsl': 'application/xml',
    '.xslfo': 'text/x-xslfo',
    '.xslt': 'application/xml',
    '.xspf': 'application/xspf+xml',
    '.xul': 'application/vnd.mozilla.xul+xml',
    '.xwd': 'image/x-xwindowdump',
    '.xyz': 'chemical/x-pdb',
    '.xz': 'application/x-xz',
    '.w2p': 'application/w2p',
    '.z': 'application/x-compress',
    '.zabw': 'application/x-abiword',
    '.zip': 'application/zip',
    '.zoo': 'application/x-zoo',
}


def contenttype(filename, default='text/plain'):
    """
    Returns the Content-Type string matching extension of the given filename.
    """

    i = filename.rfind('.')
    if i >= 0:
        default = CONTENT_TYPE.get(filename[i:].lower(), default)
        j = filename.rfind('.', 0, i)
        if j >= 0:
            default = CONTENT_TYPE.get(filename[j:].lower(), default)
    if default.startswith('text/'):
        default += '; charset=utf-8'
    return default



# -*- coding: utf-8 -*-
"""
    emmett.serializers
    ------------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from functools import partial
from typing import Any, Callable, Dict, Union

from .html import tag, htmlescape

try:
    import orjson
    _json_impl = orjson.dumps
    _json_opts = {
        "option": orjson.OPT_NON_STR_KEYS | orjson.OPT_NAIVE_UTC
    }
    _json_type = "bytes"
except ImportError:
    import rapidjson
    _json_impl = rapidjson.dumps
    _json_opts = {
        "datetime_mode": rapidjson.DM_ISO8601 | rapidjson.DM_NAIVE_IS_UTC,
        "number_mode": rapidjson.NM_NATIVE | rapidjson.NM_DECIMAL
    }
    _json_type = "str"

_json_safe_table = {
    'u2028': [r'\u2028', '\\u2028'],
    'u2029': [r'\u2029', '\\u2029']
}


class Serializers:
    _registry_: Dict[str, Callable[[Any], Union[bytes, str]]] = {}

    @classmethod
    def register_for(cls, target):
        def wrap(f):
            cls._registry_[target] = f
            return f
        return wrap

    @classmethod
    def get_for(cls, target):
        return cls._registry_[target]


def _json_default(obj):
    if hasattr(obj, '__json__'):
        return obj.__json__()
    raise TypeError


json = partial(
    _json_impl,
    default=_json_default,
    **_json_opts
)


def json_safe(value):
    rv = json(value)
    for val, rep in _json_safe_table.values():
        rv.replace(val, rep)
    return rv


def xml_encode(value, key=None, quote=True):
    if hasattr(value, '__xml__'):
        return value.__xml__(key, quote)
    if isinstance(value, dict):
        return tag[key](
            *[
                tag[k](xml_encode(v, None, quote))
                for k, v in value.items()
            ])
    if isinstance(value, list):
        return tag[key](
            *[
                tag[item](xml_encode(item, None, quote))
                for item in value
            ])
    return htmlescape(value)


Serializers.register_for('json')(json)


@Serializers.register_for('xml')
def xml(value, encoding='UTF-8', key='document', quote=True):
    rv = ('<?xml version="1.0" encoding="%s"?>' % encoding) + \
        str(xml_encode(value, key, quote))
    return rv



# -*- coding: utf-8 -*-
"""
    emmett.datastructures
    ---------------------

    Provide some useful data structures.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import copy
import hashlib
import pickle

from typing import Dict, Optional

from ._internal import ImmutableList
from .typing import KT, VT


class sdict(Dict[KT, VT]):
    #: like a dictionary except `obj.foo` can be used in addition to
    #  `obj['foo']`, and setting obj.foo = None deletes item foo.
    __slots__ = ()

    __setattr__ = dict.__setitem__  # type: ignore
    __delattr__ = dict.__delitem__  # type: ignore
    __getitem__ = dict.get  # type: ignore

    # see http://stackoverflow.com/questions/10364332/how-to-pickle-python-object-derived-from-dict
    def __getattr__(self, key: str) -> Optional[VT]:
        if key.startswith('__'):
            raise AttributeError
        return self.get(key, None)  # type: ignore

    __repr__ = lambda self: '<sdict %s>' % dict.__repr__(self)
    __getstate__ = lambda self: None
    __copy__ = lambda self: sdict(self)
    __deepcopy__ = lambda self, memo: sdict(copy.deepcopy(dict(self)))


class ConfigData(sdict[KT, VT]):
    #: like sdict, except it autogrows creating sub-sdict attributes.
    #  Useful for configurations.
    __slots__ = ()

    def __getitem__(self, key):
        if key not in self.keys():
            self[key] = sdict()
        return super().__getitem__(key)

    __getattr__ = __getitem__


class SessionData(sdict):
    __slots__ = ('__sid', '__hash', '__expires', '__dump')

    def __init__(self, initial=None, sid=None, expires=None):
        sdict.__init__(self, initial or ())
        object.__setattr__(
            self, '_SessionData__dump', pickle.dumps(sdict(self)))
        h = hashlib.md5(self._dump).hexdigest()
        object.__setattr__(self, '_SessionData__sid', sid)
        object.__setattr__(self, '_SessionData__hash', h)
        object.__setattr__(self, '_SessionData__expires', expires)

    @property
    def _sid(self):
        return self.__sid

    @property
    def _modified(self):
        dump = pickle.dumps(sdict(self))
        h = hashlib.md5(dump).hexdigest()
        if h != self.__hash:
            object.__setattr__(self, '_SessionData__dump', dump)
            return True
        return False

    @property
    def _expiration(self):
        return self.__expires

    @property
    def _dump(self):
        # note: self.__dump is updated only on _modified call
        return self.__dump

    def _expires_after(self, value):
        object.__setattr__(self, '_SessionData__expires', value)


def _unique_list(seq, hashfunc=None):
    seen = set()
    seen_add = seen.add
    if not hashfunc:
        return [x for x in seq if x not in seen and not seen_add(x)]
    return [
        x for x in seq if hashfunc(x) not in seen and not seen_add(hashfunc(x))
    ]


class OrderedSet(set):
    def __init__(self, d=None):
        set.__init__(self)
        self._list = []
        if d is not None:
            self._list = _unique_list(d)
            set.update(self, self._list)
        else:
            self._list = []

    def add(self, element):
        if element not in self:
            self._list.append(element)
        set.add(self, element)

    def remove(self, element):
        set.remove(self, element)
        self._list.remove(element)

    def insert(self, pos, element):
        if element not in self:
            self._list.insert(pos, element)
        set.add(self, element)

    def discard(self, element):
        if element in self:
            self._list.remove(element)
            set.remove(self, element)

    def clear(self):
        set.clear(self)
        self._list = []

    def __getitem__(self, key):
        return self._list[key]

    def __iter__(self):
        return iter(self._list)

    def __add__(self, other):
        return self.union(other)

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self._list)

    __str__ = __repr__

    def update(self, iterable):
        for e in iterable:
            if e not in self:
                self._list.append(e)
                set.add(self, e)
        return self

    __ior__ = update

    def union(self, other):
        result = self.__class__(self)
        result.update(other)
        return result

    __or__ = union

    def intersection(self, other):
        other = set(other)
        return self.__class__(a for a in self if a in other)

    __and__ = intersection

    def symmetric_difference(self, other):
        other = set(other)
        result = self.__class__(a for a in self if a not in other)
        result.update(a for a in other if a not in self)
        return result

    __xor__ = symmetric_difference

    def difference(self, other):
        other = set(other)
        return self.__class__(a for a in self if a not in other)

    __sub__ = difference

    def intersection_update(self, other):
        other = set(other)
        set.intersection_update(self, other)
        self._list = [a for a in self._list if a in other]
        return self

    __iand__ = intersection_update

    def symmetric_difference_update(self, other):
        set.symmetric_difference_update(self, other)
        self._list = [a for a in self._list if a in self]
        self._list += [a for a in other._list if a in self]
        return self

    __ixor__ = symmetric_difference_update

    def difference_update(self, other):
        set.difference_update(self, other)
        self._list = [a for a in self._list if a in self]
        return self

    __isub__ = difference_update


class Accept(ImmutableList):
    def __init__(self, values=()):
        if values is None:
            list.__init__(self)
            self.provided = False
        elif isinstance(values, Accept):
            self.provided = values.provided
            list.__init__(self, values)
        else:
            self.provided = True
            values = sorted(values, key=lambda x: (x[1], x[0]), reverse=True)
            list.__init__(self, values)

    def _value_matches(self, value, item):
        return item == '*' or item.lower() == value.lower()

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.quality(key)
        return list.__getitem__(self, key)

    def quality(self, key):
        for item, quality in self:
            if self._value_matches(key, item):
                return quality
        return 0

    def __contains__(self, value):
        for item, quality in self:
            if self._value_matches(value, item):
                return True
        return False

    def __repr__(self):
        return '%s([%s])' % (
            self.__class__.__name__,
            ', '.join('(%r, %s)' % (x, y) for x, y in self)
        )

    def index(self, key):
        if isinstance(key, str):
            for idx, (item, quality) in enumerate(self):
                if self._value_matches(key, item):
                    return idx
            raise ValueError(key)
        return list.index(self, key)

    def find(self, key):
        try:
            return self.index(key)
        except ValueError:
            return -1

    def values(self):
        for item in self:
            yield item[0]

    def to_header(self):
        result = []
        for value, quality in self:
            if quality != 1:
                value = '%s;q=%s' % (value, quality)
            result.append(value)
        return ','.join(result)

    def __str__(self):
        return self.to_header()

    def best_match(self, matches, default=None):
        best_quality = -1
        result = default
        for server_item in matches:
            for client_item, quality in self:
                if quality <= best_quality:
                    break
                if (
                    self._value_matches(server_item, client_item) and
                    quality > 0
                ):
                    best_quality = quality
                    result = server_item
        return result

    @property
    def best(self):
        if self:
            return self[0][0]



# -*- coding: utf-8 -*-
"""
    emmett.validators.inside
    ------------------------

    Validators that check presence/absence of given value in a set.

    :copyright: 2014 Giovanni Barillari

    Based on the web2py's validators (http://www.web2py.com)
    :copyright: (c) by Massimo Di Pierro <mdipierro@cs.depaul.edu>

    :license: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)
"""

# TODO: check unicode conversions
from .._shortcuts import to_unicode
from ..ctx import current
from ..utils import cachedprop
from .basic import Validator
from .helpers import options_sorter, translate


class inRange(Validator):
    def __init__(self, minimum=None, maximum=None, include=(True, False), message=None):
        super().__init__(message=message)
        self.minimum = minimum
        self.maximum = maximum
        self.inc = include

    def _gt(self, val1, val2, eq=False):
        if eq:
            return val1 >= val2
        return val1 > val2

    def _lt(self, val1, val2, eq=False):
        if eq:
            return val1 <= val2
        return val1 < val2

    def __call__(self, value):
        minimum = self.minimum() if callable(self.minimum) else self.minimum
        maximum = self.maximum() if callable(self.maximum) else self.maximum
        if (
            (minimum is None or self._gt(value, minimum, self.inc[0])) and
            (maximum is None or self._lt(value, maximum, self.inc[1]))
        ):
            return value, None
        return value, translate(
            self._range_error(self.message, minimum, maximum)
        )

    def _range_error(self, message, minimum, maximum):
        if message is None:
            message = "Enter a value"
            if minimum is not None and maximum is not None:
                message += " between {min} and {max}"
            elif minimum is not None:
                message += " greater than or equal to {min}"
            elif maximum is not None:
                message += " less than or equal to {max}"
        if isinstance(maximum, int):
            maximum -= 1
        return translate(message).format(min=minimum, max=maximum)


class inSet(Validator):
    def __init__(
        self,
        theset,
        labels=None,
        multiple=False,
        zero=None,
        sort=False,
        message=None
    ):
        super().__init__(message=message)
        self.multiple = multiple
        if (
            theset and
            isinstance(theset, (tuple, list)) and
            isinstance(theset[0], (tuple, list)) and
            len(theset[0]) == 2
        ):
            lset, llabels = [], []
            for item, label in theset:
                lset.append(str(item))
                llabels.append(str(label))
            self.theset = lset
            self.labels = llabels
        elif (
            theset and
            isinstance(theset, dict)
        ):
            lset, llabels = [], []
            for item, label in theset.items():
                lset.append(str(item))
                llabels.append(str(label))
            self.theset = lset
            self.labels = llabels
        else:
            self.theset = [str(item) for item in theset]
            self.labels = labels
        self.zero = zero
        self.sort = sort

    def options(self, zero=True):
        if not self.labels:
            items = [(k, k) for (i, k) in enumerate(self.theset)]
        else:
            items = [(k, self.labels[i]) for (i, k) in enumerate(self.theset)]
        if self.sort:
            items.sort(options_sorter)
        if zero and self.zero is not None and not self.multiple:
            items.insert(0, ('', self.zero))
        return items

    def __call__(self, value):
        if self.multiple:
            if not value:
                values = []
            elif isinstance(value, (tuple, list)):
                values = value
            else:
                values = [value]
        else:
            values = [value]
        failures = [
            x for x in values
            if (to_unicode(x) or '') not in self.theset]
        if failures and self.theset:
            if self.multiple and (value is None or value == ''):
                return ([], None)
            return value, translate(self.message)
        if self.multiple:
            if (
                isinstance(self.multiple, (tuple, list)) and
                not self.multiple[0] <= len(values) < self.multiple[1]
            ):
                return values, translate(self.message)
            return values, None
        return value, None


class DBValidator(Validator):
    def __init__(self, db, tablename, fieldname='id', dbset=None, message=None):
        super().__init__(message=message)
        self.db = db
        self.tablename = tablename
        self.fieldname = fieldname
        self._dbset = dbset

    def __call__(self, value):
        raise NotImplementedError

    @cachedprop
    def table(self):
        return self.db[self.tablename]

    @cachedprop
    def dbset(self):
        if self._dbset:
            return self._dbset(self.db)
        return self.db(self.table)

    @cachedprop
    def field(self):
        return self.table[self.fieldname]


class inDB(DBValidator):
    def __init__(
        self,
        db,
        tablename,
        fieldname='id',
        dbset=None,
        label_field=None,
        multiple=False,
        orderby=None,
        message=None
    ):
        super().__init__(
            db,
            tablename,
            fieldname=fieldname,
            dbset=dbset,
            message=message
        )
        self.label_field = label_field
        self.multiple = multiple
        self.orderby = orderby

    @cachedprop
    def sorting(self):
        if callable(self.orderby):
            return self.orderby(self.table)
        return None

    def _get_rows(self):
        return self.dbset.select(orderby=self.sorting)

    def options(self, zero=True):
        records = self._get_rows()
        if self.label_field:
            items = [(r.id, str(r[self.label_field])) for r in records]
        elif self.db[self.tablename]._format:
            items = [(r.id, self.db[self.tablename]._format % r) for r in records]
        else:
            items = [(r.id, r.id) for r in records]
        #if self.sort:
        #    items.sort(options_sorter)
        #if zero and self.zero is not None and not self.multiple:
        #    items.insert(0, ('', self.zero))
        return items

    def __call__(self, value):
        if self.multiple:
            values = value if isinstance(value, list) else [value]
            records = self.dbset.where(
                self.field.belongs(values)
            ).select(self.field, distinct=True).column(self.field)
            if set(values).issubset(set(records)):
                return values, None
        else:
            if self.dbset.where(self.field == value).count():
                return value, None
        return value, translate(self.message)


class notInDB(DBValidator):
    def __call__(self, value):
        row = self.dbset.where(
            self.field == value
        ).select(limitby=(0, 1)).first()
        if row:
            record_id = getattr(current, '_dbvalidation_record_id_', None)
            if row.id != record_id:
                return value, translate(self.message)
        return value, None



# -*- coding: utf-8 -*-
"""
    emmett.asgi.loops.uvloop
    ------------------------

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import asyncio

from . import loops


@loops.register('uvloop', packages=['uvloop'])
def build_uv_loop(uvloop):
    asyncio.get_event_loop().close()
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
    return asyncio.get_event_loop()



# -*- coding: utf-8 -*-
"""
    emmett.wrappers
    ---------------

    Provides request and response wrappers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import re

from http.cookies import SimpleCookie
from typing import Any, List, Type, TypeVar, Union
from urllib.parse import parse_qs

from ..asgi.typing import Scope, Receive, Send
from ..datastructures import Accept, sdict
from ..language.helpers import LanguageAccept
from ..typing import T
from ..utils import cachedprop
from .helpers import Headers

AcceptType = TypeVar("AcceptType", bound=Accept)

_regex_accept = re.compile(r'''
    ([^\s;,]+(?:[ \t]*;[ \t]*(?:[^\s;,q][^\s;,]*|q[^\s;,=][^\s;,]*))*)
    (?:[ \t]*;[ \t]*q=(\d*(?:\.\d+)?)[^,]*)?''', re.VERBOSE)


class Wrapper:
    def __getitem__(self, name: str) -> Any:
        return getattr(self, name, None)

    def __setitem__(self, name: str, value: Any):
        setattr(self, name, value)


class ScopeWrapper(Wrapper):
    __slots__ = ('_scope', '_receive', '_send', 'scheme', 'path')

    def __init__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send
    ):
        self._scope = scope
        self._receive = receive
        self._send = send
        self.scheme: str = scope['scheme']
        self.path: str = scope['emt.path']

    def __parse_accept_header(
        self,
        value: str,
        cls: Type[AcceptType]
    ) -> AcceptType:
        if not value:
            return cls(None)
        result = []
        for match in _regex_accept.finditer(value):
            mq = match.group(2)
            if not mq:
                quality = 1.0
            else:
                quality = max(min(float(mq), 1), 0)
            result.append((match.group(1), quality))
        return cls(result)

    @cachedprop
    def headers(self) -> Headers:
        return Headers(self._scope)

    @cachedprop
    def host(self) -> str:
        return self.headers.get('host')

    @cachedprop
    def accept_language(self) -> LanguageAccept:
        return self.__parse_accept_header(
            self.headers.get('accept-language'), LanguageAccept
        )

    @cachedprop
    def cookies(self) -> SimpleCookie:
        cookies: SimpleCookie = SimpleCookie()
        for cookie in self.headers.get('cookie', '').split(';'):
            cookies.load(cookie)
        return cookies

    @cachedprop
    def query_params(self) -> sdict[str, Union[str, List[str]]]:
        rv: sdict[str, Any] = sdict()
        for key, values in parse_qs(
            self._scope['query_string'].decode('latin-1'), keep_blank_values=True
        ).items():
            if len(values) == 1:
                rv[key] = values[0]
                continue
            rv[key] = values
        return rv



# -*- coding: utf-8 -*-
"""
    emmett.tools.mailer
    -------------------

    Provides mail facilities for Emmett.

    :copyright: 2014 Giovanni Barillari

    Based on the code of flask-mail
    :copyright: (c) 2010 by Dan Jacob.

    :license: BSD-3-Clause
"""

import smtplib
import time

from contextlib import contextmanager
from email import charset as _charsetreg, policy
from email.encoders import encode_base64
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.header import Header
from email.utils import formatdate, formataddr, make_msgid, parseaddr
from functools import wraps

from ..extensions import Extension
from ..libs.contenttype import contenttype
from ..utils import cachedprop

_charsetreg.add_charset('utf-8', _charsetreg.SHORTEST, None, 'utf-8')
message_policy = policy.SMTP


def _has_newline(line):
    if line and ('\r' in line or '\n' in line):
        return True
    return False


def sanitize_subject(subject, encoding='utf-8'):
    try:
        subject.encode('ascii')
    except UnicodeEncodeError:
        subject = Header(subject, encoding).encode()
    return subject


def sanitize_address(address, encoding='utf-8'):
    if isinstance(address, str):
        address = parseaddr(address)
    name, address = address
    name = Header(name, encoding).encode()
    try:
        address.encode('ascii')
    except UnicodeEncodeError:
        if '@' in address:
            localpart, domain = address.split('@', 1)
            localpart = str(Header(localpart, encoding))
            domain = domain.encode('idna').decode('ascii')
            address = '@'.join([localpart, domain])
        else:
            address = Header(address, encoding).encode()
    return formataddr((name, address))


def sanitize_addresses(addresses, encoding='utf-8'):
    return map(lambda address: sanitize_address(address, encoding), addresses)


class MailServer(object):
    def __init__(self, ext):
        self.ext = ext
        self.config = self.ext.config

    def __enter__(self):
        self.host = self.configure_host()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.host.quit()

    def configure_host(self):
        if self.config.use_ssl:
            host = smtplib.SMTP_SSL(self.config.server, self.config.port)
        else:
            host = smtplib.SMTP(self.config.server, self.config.port)
        host.set_debuglevel(int(self.config.debug))
        if self.config.use_tls:
            host.starttls()
        if self.config.username and self.config.password:
            host.login(self.config.username, self.config.password)
        return host

    def send(self, message):
        self.host.sendmail(
            sanitize_address(message.sender),
            list(sanitize_addresses(message.all_recipients)),
            str(message).encode('utf8'),
            message.mail_options,
            message.rcpt_options)
        return True


class Attachment(object):
    def __init__(
        self, filename=None, data=None, content_type=None, disposition=None,
        headers=None
    ):
        if not content_type and filename:
            content_type = contenttype(filename).split(";")[0]
        self.filename = filename
        self.content_type = content_type or contenttype(filename).split(";")[0]
        self.data = data
        self.disposition = disposition or 'attachment'
        self.headers = headers or {}


class Mail(object):
    def __init__(
        self, ext, subject='', recipients=None, body=None, html=None,
        alts=None, sender=None, cc=None, bcc=None, attachments=None,
        reply_to=None, date=None, charset='utf-8', extra_headers=None,
        mail_options=None, rcpt_options=None
    ):
        sender = sender or ext.config.sender
        if isinstance(sender, tuple):
            sender = "%s <%s>" % sender
        self.ext = ext
        self.recipients = recipients or []
        self.subject = subject
        self.sender = sender
        self.reply_to = reply_to
        self.cc = cc or []
        self.bcc = bcc or []
        self.body = body
        self.alts = dict(alts or {})
        self.html = html
        self.date = date
        self.msgId = make_msgid()
        self.charset = charset
        self.extra_headers = extra_headers
        self.mail_options = mail_options or []
        self.rcpt_options = rcpt_options or []
        self.attachments = attachments or []
        for attr in ['recipients', 'cc', 'bcc']:
            if not isinstance(getattr(self, attr), list):
                setattr(self, attr, [getattr(self, attr)])

    @property
    def all_recipients(self):
        return set(self.recipients) | set(self.bcc) | set(self.cc)

    @property
    def html(self):
        return self.alts.get('html')

    @html.setter
    def html(self, value):
        if value is None:
            self.alts.pop('html', None)
        else:
            self.alts['html'] = value

    def has_bad_headers(self):
        headers = [self.sender, self.reply_to] + self.recipients
        for header in headers:
            if _has_newline(header):
                return True
        if self.subject:
            if _has_newline(self.subject):
                for linenum, line in enumerate(self.subject.split('\r\n')):
                    if not line:
                        return True
                    if linenum > 0 and line[0] not in '\t ':
                        return True
                    if _has_newline(line):
                        return True
                    if len(line.strip()) == 0:
                        return True
        return False

    def _mimetext(self, text, subtype='plain'):
        return MIMEText(text, _subtype=subtype, _charset=self.charset)

    @cachedprop
    def message(self):
        attachments = self.attachments or []
        if len(attachments) == 0 and not self.alts:
            # No html content and zero attachments means plain text
            msg = self._mimetext(self.body)
        elif len(attachments) > 0 and not self.alts:
            # No html and at least one attachment means multipart
            msg = MIMEMultipart()
            msg.attach(self._mimetext(self.body))
        else:
            # Anything else
            msg = MIMEMultipart()
            alternative = MIMEMultipart('alternative')
            alternative.attach(self._mimetext(self.body, 'plain'))
            for mimetype, content in self.alts.items():
                alternative.attach(self._mimetext(content, mimetype))
            msg.attach(alternative)
        if self.subject:
            msg['Subject'] = sanitize_subject(self.subject, self.charset)
        msg['From'] = sanitize_address(self.sender, self.charset)
        msg['To'] = ', '.join(
            list(set(sanitize_addresses(self.recipients, self.charset))))
        msg['Date'] = formatdate(self.date, localtime=True)
        msg['Message-ID'] = self.msgId
        if self.cc:
            msg['Cc'] = ', '.join(
                list(set(sanitize_addresses(self.cc, self.charset))))
        if self.reply_to:
            msg['Reply-To'] = sanitize_address(self.reply_to, self.charset)
        if self.extra_headers:
            for k, v in self.extra_headers.items():
                msg[k] = v
        for attachment in attachments:
            f = MIMEBase(*attachment.content_type.split('/'))
            f.set_payload(attachment.data)
            encode_base64(f)
            filename = attachment.filename
            try:
                filename and filename.encode('ascii')
            except UnicodeEncodeError:
                filename = ('UTF8', '', filename)
            f.add_header(
                'Content-Disposition', attachment.disposition,
                filename=filename)
            for key, value in attachment.headers.items():
                f.add_header(key, value)
            msg.attach(f)
        if message_policy:
            msg.policy = message_policy
        return msg

    def __str__(self):
        return self.message.as_string()

    def send(self):
        return self.ext.send(self)

    def add_recipient(self, recipient):
        self.recipients.append(recipient)

    def attach(
        self, filename=None, data=None, content_type=None, disposition=None,
        headers=None
    ):
        self.attachments.append(
            Attachment(filename, data, content_type, disposition, headers))


class MailExtension(Extension):
    namespace = 'mailer'

    default_config = {
        'server': '127.0.0.1',
        'username': None,
        'password': None,
        'port': 25,
        'use_tls': False,
        'use_ssl': False,
        'sender': None,
        'debug': False,
        'suppress': False
    }

    def on_load(self):
        self.after_send = []
        self.dispatch = self._send

    def _send(self, message):
        with self.server() as server:
            return server.send(message)

    def mail(self, *args, **kwargs):
        return Mail(self, *args, **kwargs)

    def send(self, message):
        assert message.all_recipients, "No recipients have been added"
        assert message.sender, "The message does not specify a sender"
        if message.has_bad_headers():
            raise RuntimeError("Bad headers in message")
        if message.date is None:
            message.date = time.time()
        if not self.config.suppress:
            rv = self.dispatch(message)
        else:
            rv = True
        for callback in self.after_send:
            callback(message)
        return rv

    def send_mail(self, *args, **kwargs):
        return self.send(Mail(self, *args, **kwargs))

    def server(self):
        return MailServer(self)


class Mailer(object):
    def __init__(self, app):
        app.use_extension(MailExtension)
        self.ext = app.ext.MailExtension

    def mail(self, *args, **kwargs):
        return self.ext.mail(*args, **kwargs)

    def send(self, message):
        return self.ext.send(message)

    def send_mail(self, *args, **kwargs):
        return self.ext.send_mail(*args, **kwargs)

    def dispatcher(self, f):
        self.ext.dispatch = _wrap_dispatcher(f)
        return f

    @contextmanager
    def store_mails(self, suppress=True):
        outbox = []

        def _record(message):
            outbox.append(message)

        suppress = suppress and suppress != self.ext.config.suppress
        if suppress:
            _last_suppress = self.ext.config.suppress
            self.ext.config.suppress = True

        self.ext.after_send.append(_record)
        try:
            yield outbox
        finally:
            if suppress:
                self.ext.config.suppress = _last_suppress
            self.ext.after_send.pop()


def _wrap_dispatcher(dispatcher):
    @wraps(dispatcher)
    def wrapped(ext, *args, **kwargs):
        return dispatcher(*args, **kwargs)
    return wrapped



# -*- coding: utf-8 -*-
"""
    emmett.orm.engines.postgres
    ---------------------------

    Provides ORM PostgreSQL engine specific features.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from pydal.adapters.postgres import (
    PostgreBoolean,
    PostgrePsycoBoolean,
    PostgrePG8000Boolean
)
from pydal.dialects import register_expression, sqltype_for
from pydal.dialects.postgre import PostgreDialectBooleanJSON
from pydal.helpers.serializers import serializers
from pydal.objects import Expression, Query
from pydal.parsers import for_type as parse_type
from pydal.parsers.postgre import PostgreBooleanAutoJSONParser
from pydal.representers import for_type as repr_type
from pydal.representers.postgre import PostgreArraysRepresenter

from . import adapters


class JSONBPostgreDialect(PostgreDialectBooleanJSON):
    @sqltype_for('jsonb')
    def type_jsonb(self):
        return 'jsonb'

    def _jcontains(self, field, data, query_env={}):
        return '(%s @> %s)' % (
            self.expand(field, query_env=query_env),
            self.expand(data, field.type, query_env=query_env)
        )

    @register_expression('jcontains')
    def _jcontains_expr(self, expr, data):
        return Query(expr.db, self._jcontains, expr, data)

    def _jin(self, field, data, query_env={}):
        return '(%s <@ %s)' % (
            self.expand(field, query_env=query_env),
            self.expand(data, field.type, query_env=query_env)
        )

    @register_expression('jin')
    def _jin_expr(self, expr, data):
        return Query(expr.db, self._jin, expr, data)

    def _jget_common(self, op, field, data, query_env):
        if not isinstance(data, int):
            _dtype = field.type if isinstance(data, (dict, list)) else 'string'
            data = self.expand(data, field_type=_dtype, query_env=query_env)
        return '%s %s %s' % (
            self.expand(field, query_env=query_env),
            op,
            str(data)
        )

    def _jget(self, field, data, query_env={}):
        return self._jget_common('->', field, data, query_env=query_env)

    @register_expression('jget')
    def _jget_expr(self, expr, data):
        return Expression(expr.db, self._jget, expr, data, expr.type)

    def _jgetv(self, field, data, query_env={}):
        return self._jget_common('->>', field, data, query_env=query_env)

    @register_expression('jgetv')
    def _jgetv_expr(self, expr, data):
        return Expression(expr.db, self._jgetv, expr, data, 'string')

    def _jpath(self, field, data, query_env={}):
        return '%s #> %s' % (
            self.expand(field, query_env=query_env),
            self.expand(data, field_type='string', query_env=query_env)
        )

    @register_expression('jpath')
    def _jpath_expr(self, expr, data):
        return Expression(expr.db, self._jpath, expr, data, expr.type)

    def _jpathv(self, field, data, query_env={}):
        return '%s #>> %s' % (
            self.expand(field, query_env=query_env),
            self.expand(data, field_type='string', query_env=query_env)
        )

    @register_expression('jpathv')
    def _jpathv_expr(self, expr, data):
        return Expression(expr.db, self._jpathv, expr, data, 'string')

    def _jhas(self, field, data, all=False, query_env={}):
        _op, _ftype = '?', 'string'
        if isinstance(data, list):
            _op = '?&' if all else '?|'
            _ftype = 'list:string'
        return '%s %s %s' % (
            self.expand(field, query_env=query_env),
            _op,
            self.expand(data, field_type=_ftype, query_env=query_env)
        )

    @register_expression('jhas')
    def _jhas_expr(self, expr, data, all=False):
        return Query(expr.db, self._jhas, expr, data, all=all)


class JSONBPostgreParser(PostgreBooleanAutoJSONParser):
    @parse_type('jsonb')
    def _jsonb(self, value):
        return value


class JSONBPostgreRepresenter(PostgreArraysRepresenter):
    @repr_type('jsonb')
    def _jsonb(self, value):
        return serializers.json(value)


@adapters.register_for('postgres')
class PostgresAdapter(PostgreBoolean):
    def _load_dependencies(self):
        super()._load_dependencies()
        self.dialect = JSONBPostgreDialect(self)
        self.parser = JSONBPostgreParser(self)
        self.representer = JSONBPostgreRepresenter(self)

    def _config_json(self):
        pass

    def _mock_reconnect(self):
        pass


@adapters.register_for('postgres:psycopg2')
class PostgresPsycoPG2Adapter(PostgrePsycoBoolean):
    def _load_dependencies(self):
        super()._load_dependencies()
        self.dialect = JSONBPostgreDialect(self)
        self.parser = JSONBPostgreParser(self)
        self.representer = JSONBPostgreRepresenter(self)

    def _config_json(self):
        pass

    def _mock_reconnect(self):
        pass


@adapters.register_for('postgres:pg8000')
class PostgresPG8000Adapter(PostgrePG8000Boolean):
    def _load_dependencies(self):
        super()._load_dependencies()
        self.dialect = JSONBPostgreDialect(self)
        self.parser = JSONBPostgreParser(self)
        self.representer = JSONBPostgreRepresenter(self)

    def _config_json(self):
        pass

    def _mock_reconnect(self):
        pass



# -*- coding: utf-8 -*-
"""
    emmett.forms
    ------------

    Provides classes to create and style forms.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Type, Union

from .ctx import current
from .datastructures import sdict
from .html import HtmlTag, tag, cat, asis
from .orm import Field, Model
from .orm.objects import Row, Table
from .security import CSRFStorage
from .utils import cachedprop
from .validators import isEmptyOr
from .wrappers.helpers import FileStorage

__all__ = ["Form", "ModelForm"]


class BaseForm(HtmlTag):
    @staticmethod
    def _get_default_style():
        return current.app.config.ui.forms_style or FormStyle

    def __init__(
        self,
        fields: List[Field],
        writable_fields: List[Field],
        csrf: Union[str, bool] = "auto",
        id_prefix: str = "",
        formstyle: Optional[Type[FormStyle]] = None,
        keepvalues: bool = False,
        onvalidation: Optional[Callable[[Form], None]] = None,
        submit: str = "Submit",
        upload: Optional[str] = None,
        _action: str = "",
        _enctype: str = "multipart/form-data",
        _method: str = "POST",
        **kwargs: Any
    ):
        self._action = _action
        self._csrf = csrf
        self._enctype = _enctype
        self._id_prefix = id_prefix
        self._formstyle = formstyle or self._get_default_style()
        self._submit_method = _method
        self._submit_text = submit
        self._upload = upload
        self.fields = fields
        self.keepvalues = keepvalues
        self.onvalidation = onvalidation
        self.writable_fields = writable_fields
        if not issubclass(self._formstyle, FormStyle):
            raise RuntimeError(
                "{!r} is an invalid Emmett form style".format(formstyle)
            )
        self._preprocess(**kwargs)

    def _preprocess(self, **kwargs):
        #: process attributes
        self.attributes = {
            "_action": self._action,
            "_enctype": self._enctype,
            "_method": self._submit_method,
            "id_prefix": self._id_prefix,
            "hidden": {},
            "submit": self._submit_text,
            "upload": self._upload
        }
        self.attributes.update(kwargs)
        #: init the form
        self._awaited = False
        self.input_params: sdict[str, Any] = sdict()
        self.errors: sdict[str, str] = sdict()
        self.params: sdict[str, Any] = sdict()
        self.files: sdict[str, FileStorage] = sdict()
        self.processed = False
        self.accepted = False
        self.formkey = "undef"

    @property
    def csrf(self) -> bool:
        return self._csrf is True or (
            self._csrf == "auto" and self._submit_method == "POST"
        )

    def _load_csrf(self):
        if not self.csrf:
            return
        if not hasattr(current, "session"):
            raise RuntimeError("You need sessions to use csrf in forms.")
        current.session._csrf = current.session._csrf or CSRFStorage()

    @property
    def _submitted(self):
        if self.csrf:
            return self.input_params._csrf_token in current.session._csrf
        return self.input_params._csrf_token == "undef"

    def _get_input_val(self, field):
        if field.type == "boolean":
            v = self.input_params.get(field.name, False)
            if v is not False:
                v = True
        elif field.type == "upload":
            v = self.input_files.get(field.name)
        else:
            v = self.input_params.get(field.name)
        return v

    def _validate_value(self, field, value):
        value, error = field.validate(value)
        if error:
            self.errors[field.name] = error
        elif field.type == "upload":
            self.files[field.name] = value
        else:
            self.params[field.name] = value

    async def _awaited_wrap(self):
        return self

    def __await__(self):
        if self._awaited:
            return self._awaited_wrap().__await__()
        self._awaited = True
        return self._process().__await__()

    async def _load_input_params(self):
        if self._submit_method == "POST":
            params = await current.request.body_params
        else:
            params = current.request.query_params
        return sdict(params)

    async def _load_input_files(self):
        if self._submit_method == "POST":
            rv = await current.request.files
        else:
            rv = sdict()
        return rv

    async def _process(self):
        self._load_csrf()
        self.input_params = await self._load_input_params()
        self.input_files = await self._load_input_files()
        # run processing if needed
        if self._submitted:
            self.processed = True
            # validate input
            for field in self.writable_fields:
                value = self._get_input_val(field)
                self._validate_value(field, value)
            # custom validation
            if not self.errors and callable(self.onvalidation):
                self.onvalidation(self)
            # end of validation
            if not self.errors:
                self.accepted = True
                if self.csrf:
                    del current.session._csrf[self.input_params._csrf_token]
        # CSRF protection logic
        if self.csrf and not self.accepted:
            self.formkey = current.session._csrf.gen_token()
        # reset default values in form
        if not self.processed or (self.accepted and not self.keepvalues):
            for field in self.fields:
                default_value = field.default() if callable(field.default) \
                    else field.default
                self.input_params[field.name] = default_value
        return self

    def _render(self):
        styler = self._formstyle(self.attributes)
        styler.on_start()
        for field in self.fields:
            value = self.input_params.get(field.name)
            error = self.errors.get(field.name)
            styler._proc_element(field, value, error)
        styler.add_buttons()
        styler._add_formkey(self.formkey)
        for key, value in self.attributes["hidden"].items():
            styler._add_hidden(key, value)
        return styler.render()

    @cachedprop
    def custom(self):
        # init
        custom = sdict()
        custom.dspval = sdict()
        custom.inpval = sdict()
        custom.label = sdict()
        custom.comment = sdict()
        custom.widget = sdict()
        # load selected styler
        styler = self._formstyle(self.attributes)
        styler.on_start()
        # load data
        for field in self.fields:
            value = self.input_params.get(field.name)
            custom.dspval[field.name] = self.input_params[field.name]
            custom.inpval[field.name] = self.input_params[field.name]
            custom.label[field.name] = field.label
            custom.comment[field.name] = field.comment
            widget, wfield = styler._get_widget(field, value)
            if not wfield:
                styler.style_widget(widget)
            custom.widget[field.name] = widget
        # add submit
        custom.submit = tag.input(_type="submit", value=self.attributes["submit"])
        # provides begin attribute
        custom.begin = asis(f"<form {self._build_html_attributes()}>")
        # add hidden stuffs to get process working
        hidden = cat()
        hidden.append(
            tag.input(_name="_csrf_token", _type="hidden", _value=self.formkey)
        )
        for key, value in self.attributes["hidden"].items():
            hidden.append(tag.input(_name=key, _type="hidden", _value=value)
        )
        # provides end attribute
        custom.end = asis(f"{hidden.__html__()}</form>")
        return custom

    def __html__(self):
        return self._render().__html__()


class Form(BaseForm):
    def __init__(
        self,
        fields: Optional[Dict[str, Field]] = None,
        csrf: Union[str, bool] = "auto",
        id_prefix: str = "",
        formstyle: Optional[Type[FormStyle]] = None,
        keepvalues: bool = False,
        onvalidation: Optional[Callable[[Form], None]] = None,
        submit: str = "Submit",
        upload: Optional[str] = None,
        _action: str = "",
        _enctype: str = "multipart/form-data",
        _method: str = "POST",
        **kwargs: Any
    ):
        fields = fields or {}
        #: get fields from kwargs
        for name, parameter in kwargs.items():
            if isinstance(parameter, Field):
                fields[name] = parameter
        for name in fields.keys():
            if name in kwargs:
                del kwargs[name]
        #: order fields correctly
        sorted_fields = []
        for name, field in fields.items():
            sorted_fields.append((name, field))
        sorted_fields.sort(key=lambda tup: tup[1]._inst_count_)
        #: init fields
        fields_list_all = []
        fields_list_writable = []
        for name, obj in sorted_fields:
            field_obj = obj._make_field(name)
            fields_list_all.append(field_obj)
            if field_obj.writable:
                fields_list_writable.append(field_obj)
        super().__init__(
            fields=fields_list_all,
            writable_fields=fields_list_writable,
            csrf=csrf,
            id_prefix=id_prefix,
            formstyle=formstyle,
            keepvalues=keepvalues,
            onvalidation=onvalidation,
            submit=submit,
            upload=upload,
            _action=_action,
            _enctype=_enctype,
            _method=_method
        )


class ModelForm(BaseForm):
    def __init__(
        self,
        table: Table,
        record: Optional[Row] = None,
        record_id: Optional[int] = None,
        fields: Union[Dict[str, List[str]], List[str]] = None,
        exclude_fields: List[str] = [],
        csrf: Union[str, bool] = "auto",
        formstyle: Optional[Type[FormStyle]] = None,
        keepvalues: bool = False,
        onvalidation: Optional[Callable[[Form], None]] = None,
        submit: str = "Submit",
        upload: Optional[str] = None,
        _action: str = "",
        _enctype: str = "multipart/form-data",
        _method: str = "POST",
        **attributes
    ):
        self.table = table
        self.record = record or table(record_id)
        #: build fields for form
        fields_list_all = []
        fields_list_writable = []
        if fields is not None:
            #: developer has selected specific fields
            if not isinstance(fields, dict):
                fields = {'writable': fields, 'readable': fields}
            for field in table:
                if field.name not in fields['readable']:
                    continue
                fields_list_all.append(field)
                if field.name in fields['writable']:
                    fields_list_writable.append(field)
        else:
            #: use table fields
            for field in table:
                if field.name in exclude_fields:
                    continue
                if not field.readable:
                    continue
                if not self.record and not field.writable:
                    #: show readable fields only on update
                    continue
                fields_list_all.append(field)
                if field.writable:
                    fields_list_writable.append(field)
                # if field.type != 'id' and field.writable and \
                #         field.name not in exclude_fields:
                #     self.fields.append(field)
        super().__init__(
            fields=fields_list_all,
            writable_fields=fields_list_writable,
            csrf=csrf,
            id_prefix=table._tablename + "_",
            formstyle=formstyle,
            keepvalues=keepvalues,
            onvalidation=onvalidation,
            submit=submit,
            upload=upload,
            _action=_action,
            _enctype=_enctype,
            _method=_method
        )

    def _validate_value(self, field, value):
        if field.type == "upload" and self.record:
            if (
                (value == b"" or value is None) and
                not self.input_params.get(field.name + "__del", False) and
                self.record[field.name]
            ):
                return
        super()._validate_value(field, value)

    async def _process(self):
        #: send record id to validators if needed
        current._dbvalidation_record_id_ = None
        if self.record:
            current._dbvalidation_record_id_ = self.record.id
        #: load super `_process`
        # await Form._process(self)
        await super()._process()
        #: clear current and run additional operations for DAL
        del current._dbvalidation_record_id_
        if self.accepted:
            for field in self.writable_fields:
                #: handle uploads
                if field.type == "upload":
                    upload = self.files[field.name]
                    del_field = field.name + "__del"
                    if not upload.filename:
                        if self.input_params.get(del_field, False):
                            self.params[field.name] = (
                                self.table[field.name].default or ""
                            )
                            # TODO: we want to physically delete file?
                        else:
                            if self.record and self.record[field.name]:
                                self.params[field.name] = self.record[field.name]
                        continue
                    else:
                        source_file, original_filename = (
                            upload.stream, upload.filename
                        )
                    newfilename = field.store(
                        source_file, original_filename, field.uploadfolder
                    )
                    if isinstance(field.uploadfield, str):
                        self.params[field.uploadfield] = source_file.read()
                    self.params[field.name] = newfilename
            #: add default values to hidden fields if needed
            if not self.record:
                fieldnames = [field.name for field in self.writable_fields]
                for field in self.table:
                    if field.name not in fieldnames and field.compute is None:
                        if field.default is not None:
                            def_val = (
                                field.default() if callable(field.default) else
                                field.default
                            )
                            self.params[field.name] = def_val
            if self.record:
                self.record.update_record(**self.params)
            else:
                self.params.id = self.table.insert(**self.params)
        if not self.processed or (self.accepted and not self.keepvalues):
            for field in self.fields:
                if self.record:
                    self.input_params[field.name] = self.record[field.name]
                self.input_params[field.name] = field.formatter(
                    self.input_params[field.name]
                )
        elif self.processed and not self.accepted and self.record:
            for field in self.writable_fields:
                if field.type == "upload" and field.name not in self.params:
                    self.input_params[field.name] = self.record[field.name]
                    self.input_params[field.name] = field.formatter(
                        self.input_params[field.name]
                    )
        return self


class FormStyle:
    _stack = []
    parent = None

    @staticmethod
    def _field_options(field):
        validator = FormStyle._validation_woptions(field)
        return validator.options(), validator.multiple

    @staticmethod
    def widget_string(attr, field, value, _class="string", _id=None):
        return tag.input(
            _type="text",
            _name=field.name,
            _value=value if value is not None else "",
            _class=_class,
            _id=_id or field.name
        )

    @staticmethod
    def widget_text(attr, field, value, _class="text", _id=None):
        return tag.textarea(
            value or "",
            _name=field.name,
            _class=_class,
            _id=_id or field.name
        )

    @staticmethod
    def widget_int(attr, field, value, _class="int", _id=None):
        return FormStyle.widget_string(attr, field, value, _class, _id)

    @staticmethod
    def widget_bigint(attr, field, value, _class="int", _id=None):
        return FormStyle.widget_string(attr, field, value, _class, _id)

    @staticmethod
    def widget_float(attr, field, value, _class="float", _id=None):
        return FormStyle.widget_string(attr, field, value, _class, _id)

    @staticmethod
    def widget_date(attr, field, value, _class="date", _id=None):
        return FormStyle.widget_string(attr, field, value, _class, _id)

    @staticmethod
    def widget_time(attr, field, value, _class="time", _id=None):
        return FormStyle.widget_string(attr, field, value, _class, _id)

    @staticmethod
    def widget_datetime(attr, field, value, _class="datetime", _id=None):
        return FormStyle.widget_string(attr, field, value, _class, _id)

    @staticmethod
    def widget_password(attr, field, value, _class="password", _id=None):
        return tag.input(
            _type="password",
            _name=field.name,
            _value=value or "",
            _class=_class,
            _id=_id or field.name
        )

    @staticmethod
    def widget_bool(attr, field, value, _class="bool", _id=None):
        return tag.input(
            _type="checkbox",
            _name=field.name,
            _checked="checked" if value else None,
            _class=_class,
            _id=_id or field.name
        )

    @staticmethod
    def widget_select(attr, field, value, _class="", _id=None):
        def selected(k):
            return "selected" if str(value) == str(k) else None

        options, multiple = FormStyle._field_options(field)
        if multiple:
            return FormStyle.widget_multiple(
                attr, field, value, options, _class=_class, _id=_id
            )
        return tag.select(
            *[
                tag.option(n, _value=k, _selected=selected(k)) for k, n in options
            ],
            _name=field.name,
            _class=_class,
            _id=_id or field.name
        )

    @staticmethod
    def widget_multiple(attr, field, values, options, _class="", _id=None):
        def selected(k):
            return "selected" if str(k) in [str(v) for v in values] else None

        values = values or []
        return tag.select(
            *[
                tag.option(n, _value=k, _selected=selected(k)) for k, n in options
            ],
            _name=field.name,
            _class=_class,
            _multiple="multiple",
            _id=_id or field.name
        )

    #: TO-DO
    #@staticmethod
    #def widget_list(attr, field, value, _class="", _id=None):
    #    return ""

    @staticmethod
    def widget_upload(attr, field, value, _class="upload", _id=None):
        def is_image(value):
            return value.split(".")[-1].lower() in ["gif", "png", "jpg", "jpeg", "bmp"]

        def _coerce_value(value):
            if isinstance(value, str) or isinstance(value, bytes):
                return value or ""
            return ""

        elements = []
        _value = _coerce_value(value)
        download_url = attr.get("upload")
        inp = tag.input(_type="file", _name=field.name, _class=_class, _id=_id)
        elements.append(inp)
        if _value and download_url:
            if callable(download_url):
                url = download_url(value)
            else:
                url = download_url + "/" + value
            if is_image(_value):
                elements.append(tag.img(_src=url, _width="120px", _class="upload_img"))
            else:
                elements.append(tag.div(tag.a(_value, _href=url)))
            requires = field.requires or []
            # delete checkbox
            if not requires or any(isinstance(v, isEmptyOr) for v in requires):
                elements.append(
                    tag.div(
                        tag.input(
                            _type="checkbox",
                            _class="checkbox",
                            _id=_id + "__del",
                            _name=field.name + "__del",
                            _style="display: inline;"
                        ),
                        tag.label(
                            "delete",
                            _for=_id + "__del",
                            _style="margin: 4px"
                        ),
                        _style="white-space: nowrap;"
                    )
                )
        return tag.div(*elements, _class="upload_wrap")

    @staticmethod
    def widget_json(attr, field, value, _id=None):
        return FormStyle.widget_text(attr, field, value, _id=_id or field.name)

    @staticmethod
    def widget_jsonb(attr, field, value, _id=None):
        return FormStyle.widget_text(attr, field, value, _id=_id or field.name)

    @staticmethod
    def widget_radio(field, value):
        options, _ = FormStyle._field_options(field)
        return cat(*[
            tag.div(
                tag.input(
                    _id=f"{field.name}_{k}",
                    _name=field.name,
                    _value=k,
                    _type="radio",
                    _checked=("checked" if str(k) == str(value) else None)
                ),
                tag.label(n, _for=f"{field.name}_{k}"),
                _class="option_wrap"
            ) for k, n in options
        ])

    def __init__(self, attributes):
        self.attr = attributes

    @staticmethod
    def _validation_woptions(field):
        ftype = field._type.split(":")[0]
        if ftype != "bool" and field.requires:
            for v in field.requires:
                if hasattr(v, "options"):
                    return v
        return None

    #: returns the widget for the field and a boolean (True if widget is
    #  defined by user, False if it comes from styler default ones)
    def _get_widget(self, field, value):
        if field.widget:
            return field.widget(field, value), True
        widget_id = self.attr["id_prefix"] + field.name
        wtype = field._type.split(":")[0]
        if self._validation_woptions(field) is not None:
            wtype = "select"
        elif wtype.startswith("reference"):
            wtype = "int"
        elif wtype.startswith("decimal"):
            wtype = "float"
        try:
            widget = getattr(self, "widget_" + wtype)(
                self.attr, field, value, _id=widget_id
            )
            if not field.writable:
                self._disable_widget(widget)
            return widget, False
        except AttributeError:
            raise RuntimeError(
                f"Missing form widget for field {field.name} of type {wtype}"
            )

    def _disable_widget(self, widget):
        widget.attributes["_disabled"] = "disabled"

    def _proc_element(self, field, value, error):
        widget, wfield = self._get_widget(field, value)
        self._stack.append(sdict(widget=widget, _wffield=wfield))
        self._add_element(field.label, field.comment, error)
        self._stack.pop(-1)

    def _add_element(self, label, comment, error):
        # style only widgets not defined by user
        if not self.element._wffield:
            self.style_widget(self.element.widget)
        self.element.label = self.create_label(label)
        if comment:
            self.element.comment = self.create_comment(comment)
        if error:
            self.element.error = self.create_error(error)
        self.add_widget(self.element.widget)

    def _add_hidden(self, key, value):
        self.parent.append(tag.input(_name=key, _type="hidden", _value=value))

    def _add_formkey(self, key):
        self.parent.append(tag.input(_name="_csrf_token", _type="hidden", _value=key))

    @property
    def element(self):
        return self._stack[-1] if self._stack else None

    def on_start(self):
        self.parent = cat()

    def style_widget(self, widget):
        pass

    def create_label(self, label):
        return tag.label(label, _for=self.element.widget["_id"], _class="emt_label")

    def create_error(self, error):
        return tag.div(error, _class="emt_error")

    def create_comment(self, comment):
        return tag.p(comment, _class="emt_help")

    def add_widget(self, widget):
        wrapper = tag.div(widget)
        if self.element.error:
            wrapper.append(self.element.error)
        if self.element.comment:
            wrapper.append(self.element.comment)
        self.parent.append(tag.div(self.element.label, wrapper))

    def add_buttons(self):
        submit = tag.input(_type="submit", _value=self.attr["submit"])
        self.parent.append(tag.div(submit))

    def render(self):
        return tag.form(self.parent, **self.attr)


def add_form_on_model(cls):
    @wraps(cls)
    def wrapped(model, *args, **kwargs):
        return cls(model.table, *args, **kwargs)
    return wrapped


setattr(Model, "form", classmethod(add_form_on_model(ModelForm)))



# -*- coding: utf-8 -*-
"""
    emmett._reloader
    ----------------

    Provides auto-reloading utilities.

    :copyright: 2014 Giovanni Barillari

    Adapted from werkzeug code (http://werkzeug.pocoo.org)
    :copyright: (c) 2015 by Armin Ronacher.

    :license: BSD-3-Clause
"""

import multiprocessing
import os
import signal
import subprocess
import ssl
import sys
import time

from itertools import chain
from typing import Optional

import click

from ._internal import locate_app
from .asgi.server import run as _asgi_run


def _iter_module_files():
    """This iterates over all relevant Python files.  It goes through all
    loaded files from modules, all files in folders of already loaded modules
    as well as all files reachable through a package.
    """
    for module in list(sys.modules.values()):
        if module is None:
            continue
        filename = getattr(module, '__file__', None)
        if filename:
            old = None
            while not os.path.isfile(filename):
                old = filename
                filename = os.path.dirname(filename)
                if filename == old:
                    break
            else:
                # if filename[-4:] in ('.pyc', '.pyo'):
                #     filename = filename[:-1]
                yield filename


def _get_args_for_reloading():
    """Returns the executable. This contains a workaround for windows
    if the executable is incorrectly reported to not have the .exe
    extension which can cause bugs on reloading.
    """
    rv = [sys.executable]
    py_script = sys.argv[0]
    if (
        os.name == 'nt' and not os.path.exists(py_script) and
        os.path.exists(py_script + '.exe')
    ):
        py_script += '.exe'
    if (
        os.path.splitext(rv[0])[1] == '.exe' and
        os.path.splitext(py_script)[1] == '.exe'
    ):
        rv.pop(0)
    rv.append(py_script)
    rv.extend(sys.argv[1:])
    return rv


class ReloaderLoop(object):
    name: str

    # monkeypatched by testsuite. wrapping with `staticmethod` is required in
    # case time.sleep has been replaced by a non-c function (e.g. by
    # `eventlet.monkey_patch`) before we get here
    _sleep = staticmethod(time.sleep)

    def __init__(self, extra_files=None, interval=1):
        self.extra_files = set(
            os.path.abspath(x) for x in extra_files or ())
        self.interval = interval

    def run(self):
        pass

    def restart_with_reloader(self):
        """Spawn a new Python interpreter with the same arguments as this one,
        but running the reloader thread.
        """
        while 1:
            click.secho('> Restarting (%s mode)' % self.name, fg='yellow')
            args = _get_args_for_reloading()
            new_environ = os.environ.copy()
            new_environ['EMMETT_RUN_MAIN'] = 'true'

            # a weird bug on windows. sometimes unicode strings end up in the
            # environment and subprocess.call does not like this, encode them
            # to latin1 and continue.
            # if os.name == 'nt' and PY2:
            #     for key, value in iteritems(new_environ):
            #         if isinstance(value, unicode):
            #             new_environ[key] = value.encode('iso-8859-1')

            exit_code = subprocess.call(args, env=new_environ)
            if exit_code != 3:
                return exit_code

    def trigger_reload(self, process, filename):
        filename = os.path.abspath(filename)
        click.secho('> Detected change in %r, reloading' % filename, fg='cyan')
        os.kill(process.pid, signal.SIGTERM)
        process.join()
        sys.exit(3)


class StatReloaderLoop(ReloaderLoop):
    name = 'stat'

    def run(self, process):
        mtimes = {}
        while 1:
            for filename in chain(_iter_module_files(), self.extra_files):
                try:
                    mtime = os.stat(filename).st_mtime
                except OSError:
                    continue

                old_time = mtimes.get(filename)
                if old_time is None:
                    mtimes[filename] = mtime
                    continue
                elif mtime > old_time:
                    self.trigger_reload(process, filename)
            self._sleep(self.interval)


reloader_loops = {
    'stat': StatReloaderLoop,
}

reloader_loops['auto'] = reloader_loops['stat']


def run_with_reloader(
    app_target,
    host,
    port,
    loop='auto',
    proto_http='auto',
    proto_ws='auto',
    log_level=None,
    access_log=None,
    ssl_certfile: Optional[str] = None,
    ssl_keyfile: Optional[str] = None,
    ssl_cert_reqs: int = ssl.CERT_NONE,
    ssl_ca_certs: Optional[str] = None,
    extra_files=None,
    interval=1,
    reloader_type='auto'
):
    reloader = reloader_loops[reloader_type](extra_files, interval)
    signal.signal(signal.SIGTERM, lambda *args: sys.exit(0))

    try:
        if os.environ.get('EMMETT_RUN_MAIN') == 'true':
            # FIXME: find a better way to have app files in stat checker
            locate_app(*app_target)

            process = multiprocessing.Process(
                target=_asgi_run,
                args=(app_target, host, port),
                kwargs={
                    "loop": loop,
                    "proto_http": proto_http,
                    "proto_ws": proto_ws,
                    "log_level": log_level,
                    "access_log": access_log,
                    "ssl_certfile": ssl_certfile,
                    "ssl_keyfile": ssl_keyfile,
                    "ssl_cert_reqs": ssl_cert_reqs,
                    "ssl_ca_certs": ssl_ca_certs
                }
            )
            process.start()
            reloader.run(process)
        else:
            sys.exit(reloader.restart_with_reloader())
    except KeyboardInterrupt:
        pass



# -*- coding: utf-8 -*-
"""
    emmett.asgi.protocols.ws.wsproto
    --------------------------------

    Provides websocket wsproto protocol implementation

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from uvicorn.protocols.websockets.wsproto_impl import WSProtocol

from . import protocols


protocols.register("wsproto")(WSProtocol)



# -*- coding: utf-8 -*-
"""
    emmett.routing.rules
    --------------------

    Provides routing rules definition apis.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import os

from typing import Any, Callable

from ..cache import RouteCacheRule
from ..pipeline import RequestPipeline, WebsocketPipeline, Pipe
from .routes import HTTPRoute, WebsocketRoute


class RoutingRule:
    __slots__ = ['router']

    def __init__(self, router, *args, **kwargs):
        self.router = router

    @property
    def app(self):
        return self.router.app

    def build_name(self, f):
        filename = os.path.realpath(f.__code__.co_filename)
        short = filename[1 + len(self.app.root_path):].rsplit('.', 1)[0]
        if not short:
            short = filename.rsplit('.', 1)[0]
        if short == "__init__":
            short = self.app.root_path.rsplit('/', 1)[-1]
        #: allow only one naming level if name is not provided
        if len(short.split(os.sep)) > 1:
            short = short.split(os.sep)[-1]
        return '.'.join(short.split(os.sep) + [f.__name__])

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        raise NotImplementedError


class HTTPRoutingRule(RoutingRule):
    __slots__ = [
        'cache_rule',
        'f',
        'head_builder',
        'hostname',
        'methods',
        'name',
        'output_type',
        'paths',
        'pipeline_flow_close',
        'pipeline_flow_open',
        'pipeline',
        'prefix',
        'response_builder',
        'schemes',
        'template_folder',
        'template_path',
        'template'
    ]

    def __init__(
        self, router, paths=None, name=None, template=None, pipeline=None,
        injectors=None, schemes=None, hostname=None, methods=None, prefix=None,
        template_folder=None, template_path=None, cache=None, output='auto'
    ):
        super().__init__(router)
        self.name = name
        self.paths = paths
        if self.paths is None:
            self.paths = []
        if not isinstance(self.paths, (list, tuple)):
            self.paths = (self.paths,)
        self.schemes = schemes or ('http', 'https')
        if not isinstance(self.schemes, (list, tuple)):
            self.schemes = (self.schemes,)
        self.methods = methods or ('get', 'post', 'head')
        if not isinstance(self.methods, (list, tuple)):
            self.methods = (self.methods,)
        self.hostname = hostname or self.app.config.hostname_default
        if prefix:
            if not prefix.startswith('/'):
                prefix = '/' + prefix
        self.prefix = prefix
        if output not in self.router._outputs:
            raise SyntaxError(
                'Invalid output specified. Allowed values are: {}'.format(
                    ', '.join(self.router._outputs.keys())))
        self.output_type = output
        self.template = template
        self.template_folder = template_folder
        self.template_path = template_path or self.app.template_path
        self.pipeline = (
            self.router.pipeline + (pipeline or []) +
            self.router.injectors + (injectors or []))
        self.cache_rule = None
        if cache:
            if not isinstance(cache, RouteCacheRule):
                raise RuntimeError(
                    'route cache argument should be a valid caching rule')
            if any(key in self.methods for key in ['get', 'head']):
                self.cache_rule = cache
        # check pipes are indeed valid pipes
        if any(not isinstance(pipe, Pipe) for pipe in self.pipeline):
            raise RuntimeError('Invalid pipeline')

    def _make_builders(self, output_type):
        builder_cls = self.router._outputs[output_type]
        return builder_cls(self), self.router._outputs['empty'](self)

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        if not self.paths:
            self.paths.append("/" + f.__name__)
        if not self.name:
            self.name = self.build_name(f)
        # is it good?
        if self.name.endswith("."):
            self.name = self.name + f.__name__
        #
        if not self.template:
            self.template = f.__name__ + self.app.template_default_extension
        if self.template_folder:
            self.template = os.path.join(self.template_folder, self.template)
        pipeline_obj = RequestPipeline(self.pipeline)
        wrapped_f = pipeline_obj(f)
        self.pipeline_flow_open = pipeline_obj._flow_open()
        self.pipeline_flow_close = pipeline_obj._flow_close()
        self.f = wrapped_f
        output_type = pipeline_obj._output_type() or self.output_type
        self.response_builder, self.head_builder = self._make_builders(output_type)
        for idx, path in enumerate(self.paths):
            self.router.add_route(HTTPRoute(self, path, idx))
        return f


class WebsocketRoutingRule(RoutingRule):
    __slots__ = [
        'f',
        'hostname',
        'name',
        'paths',
        'pipeline_flow_close',
        'pipeline_flow_open',
        'pipeline_flow_receive',
        'pipeline_flow_send',
        'pipeline',
        'prefix',
        'schemes'
    ]

    def __init__(
        self, router, paths=None, name=None, pipeline=None, schemes=None,
        hostname=None, prefix=None
    ):
        super().__init__(router)
        self.name = name
        self.paths = paths
        if self.paths is None:
            self.paths = []
        if not isinstance(self.paths, (list, tuple)):
            self.paths = (self.paths,)
        self.schemes = schemes or ('ws', 'wss')
        if not isinstance(self.schemes, (list, tuple)):
            self.schemes = (self.schemes,)
        self.hostname = hostname or self.app.config.hostname_default
        if prefix:
            if not prefix.startswith('/'):
                prefix = '/' + prefix
        self.prefix = prefix
        self.pipeline = self.router.pipeline + (pipeline or [])
        # check pipes are indeed valid pipes
        if any(not isinstance(pipe, Pipe) for pipe in self.pipeline):
            raise RuntimeError('Invalid pipeline')

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        if not self.paths:
            self.paths.append("/" + f.__name__)
        if not self.name:
            self.name = self.build_name(f)
        # is it good?
        if self.name.endswith("."):
            self.name = self.name + f.__name__
        #
        pipeline_obj = WebsocketPipeline(self.pipeline)
        wrapped_f = pipeline_obj(f)
        self.pipeline_flow_open = pipeline_obj._flow_open()
        self.pipeline_flow_close = pipeline_obj._flow_close()
        self.pipeline_flow_receive = pipeline_obj._flow_receive()
        self.pipeline_flow_send = pipeline_obj._flow_send()
        self.f = wrapped_f
        for idx, path in enumerate(self.paths):
            self.router.add_route(WebsocketRoute(self, path, idx))
        return f



# -*- coding: utf-8 -*-
"""
    emmett.validators.consist
    -------------------------

    Validators that check the value is of a certain type.

    :copyright: 2014 Giovanni Barillari

    Based on the web2py's validators (http://www.web2py.com)
    :copyright: (c) by Massimo Di Pierro <mdipierro@cs.depaul.edu>

    :license: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)
"""

import decimal
import json
import re
import struct

from datetime import date, time, datetime, timedelta
from time import strptime
from urllib.parse import unquote as url_unquote

from ..serializers import Serializers
from ..utils import parse_datetime
from .basic import Validator, ParentValidator, _is, Matches
from .helpers import (
    _DEFAULT,
    _UTC,
    official_top_level_domains,
    official_url_schemes,
    translate,
    unofficial_url_schemes,
    url_split_regex
)

try:
    import ipaddress
except ImportError:
    ipaddress = None

_utc = _UTC()


class isInt(_is):
    rule = re.compile(r"^[+-]?\d+$")

    def check(self, value):
        return int(value), None


class isFloat(_is):
    def __init__(self, dot=".", message=None):
        super().__init__(message=message)
        self.dot = dot

    def check(self, value):
        try:
            v = float(str(value).replace(self.dot, '.'))
            return v, None
        except (ValueError, TypeError):
            pass
        return value, translate(self.message)

    def formatter(self, value):
        if value is None:
            return None
        val = str(value)
        if '.' not in val:
            val += '.00'
        else:
            val += '0' * (2 - len(val.split('.')[1]))
        return val.replace('.', self.dot)


class isDecimal(isFloat):
    def check(self, value):
        try:
            if isinstance(value, decimal.Decimal):
                v = value
            else:
                v = decimal.Decimal(str(value).replace(self.dot, '.'))
            return v, None
        except (ValueError, TypeError, decimal.InvalidOperation):
            return value, translate(self.message)


class isTime(_is):
    rule = re.compile(
        r"((?P<h>[0-9]+))([^0-9 ]+(?P<m>[0-9 ]+))?"
        r"([^0-9ap ]+(?P<s>[0-9]*))?((?P<d>[ap]m))?"
    )

    def __call__(self, value):
        return super().__call__(value.lower() if value else value)

    def check(self, value):
        val = self.rule.match(value)
        try:
            (h, m, s) = (int(val.group('h')), 0, 0)
            if not val.group('m') is None:
                m = int(val.group('m'))
            if not val.group('s') is None:
                s = int(val.group('s'))
            if val.group('d') == 'pm' and 0 < h < 12:
                h = h + 12
            if val.group('d') == 'am' and h == 12:
                h = 0
            if not (h in range(24) and m in range(60) and s
                    in range(60)):
                raise ValueError(
                    'Hours or minutes or seconds are outside of allowed range'
                )
            val = time(h, m, s)
            return val, None
        except AttributeError:
            pass
        except ValueError:
            pass
        return value, translate(self.message)


class isDate(_is):
    def __init__(self, format='%Y-%m-%d', timezone=None, message=None):
        super().__init__(message=message)
        self.format = translate(format)
        self.timezone = timezone
        self.extremes = {}

    def _parse(self, value):
        y, m, d, hh, mm, ss, t0, t1, t2 = strptime(value, str(self.format))
        return date(y, m, d)

    def _check_instance(self, value):
        return isinstance(value, date)

    def check(self, value):
        if self._check_instance(value):
            if self.timezone is not None:
                value = value - timedelta(seconds=self.timezone * 3600)
            return value, None
        try:
            val = self._parse(value)
            if self.timezone is not None:
                val = self.timezone.localize(val).astimezone(_utc)
            return val, None
        except Exception:
            self.extremes.update(isDate.nice(self.format))
            return value, translate(self.message) % self.extremes

    def _formatter_obj(self, value):
        return datetime(value.year, value.month, value.day)

    def formatter(self, value):
        if value is None:
            return None
        d = self._formatter_obj(value)
        if self.timezone is not None:
            d = d.replace(tzinfo=_utc).astimezone(self.timezone)
        return d.strftime(str(self.format))

    @staticmethod
    def nice(format):
        codes = (
            ('%Y', '1963'), ('%y', '63'), ('%d', '28'), ('%m', '08'),
            ('%b', 'Aug'), ('%B', 'August'), ('%H', '14'), ('%I', '02'),
            ('%p', 'PM'), ('%M', '30'), ('%S', '59')
        )
        for (a, b) in codes:
            format = format.replace(a, b)
        return dict(format=format)


class isDatetime(isDate):
    def __init__(self, format=_DEFAULT, **kwargs):
        self._parse, format = self._get_parser(format)
        super().__init__(format=format, **kwargs)

    def _get_parser(self, format):
        if format is _DEFAULT:
            return self._parse_pendulum, '%Y-%m-%dT%H:%M:%S'
        return self._parse_strptime, format

    def _parse_strptime(self, value):
        y, m, d, hh, mm, ss, t0, t1, t2 = strptime(value, str(self.format))
        return datetime(y, m, d, hh, mm, ss)

    def _parse_pendulum(self, value):
        return parse_datetime(value).in_timezone('UTC')

    def _check_instance(self, value):
        return isinstance(value, datetime)

    def _formatter_obj(self, value):
        return datetime(
            value.year,
            value.month,
            value.day,
            value.hour,
            value.minute,
            value.second
        )


class isEmail(_is):
    rule = re.compile(
        r"^(?!\.)([-a-z0-9!\#$%&'*+/=?^_`{|}~]|(?<!\.)\.)+(?<!\.)@"
        r"(localhost|([a-z0-9]([-\w]*[a-z0-9])?\.)+[a-z]{2,})$",
        re.VERBOSE | re.IGNORECASE
    )

    def __init__(self, banned=None, forced=None, message=None):
        super().__init__(message=message)
        self.banned = banned
        self.forced = forced

    def check(self, value):
        domain = value.split('@')[1]
        if (
            (not self.banned or not self.banned.match(domain)) and
            (not self.forced or self.forced.match(domain))
        ):
            return value, None
        return value, translate(self.message)


class isList(ParentValidator):
    def __init__(self, children, splitter=None, message=None):
        super().__init__(children, message=message)
        self.splitter = re.compile(r"[^" + splitter + r"\s]+") if splitter else None

    def __call__(self, value):
        if self.splitter is not None and isinstance(value, str):
            values = self.splitter.findall(value)
        else:
            if not isinstance(value, list):
                values = [value]
            else:
                values = value
            values = [val for val in values if str(val).strip()]
        nvals = []
        for val in values:
            v = val
            for child in self.children:
                v, e = child(v)
                if e:
                    return val, e
            nvals.append(v)
        values = nvals
        return values, None


class isJSON(_is):
    JSONErrors = (NameError, TypeError, ValueError, AttributeError, KeyError)

    def check(self, value):
        if isinstance(value, str):
            try:
                v = json.loads(value)
                return v, None
            except self.JSONErrors:
                return value, translate(self.message)
        if not isinstance(value, (dict, list)):
            return value, translate(self.message)
        try:
            Serializers.get_for('json')(value)
        except self.JSONErrors:
            return value, translate(self.message)
        return value, None

    def formatter(self, value):
        if value is None:
            return None
        return json.dumps(value)


class isAlphanumeric(Matches):
    message = "Enter only letters, numbers, and underscores"

    def __init__(self, message=None):
        super().__init__(r"^[\w]*$", message=message)


class isImage(Validator):
    # Checks if file uploaded through file input was saved in one of selected
    # image formats and has dimensions within given boundaries.
    message = "Invalid image"

    def __init__(
        self,
        extensions=('bmp', 'gif', 'jpeg', 'png'),
        width={'min': 0, 'max': 10000},
        height={'min': 0, 'max': 100000},
        message=None
    ):
        super().__init__(message=message)
        self.extensions = extensions
        self.maxsize = (width.get('max', 10000), height.get('max', 10000))
        self.minsize = (width.get('min', 0), height.get('min', 0))

    def __call__(self, value):
        try:
            extension = value.filename.rfind('.')
            assert extension >= 0
            extension = value.filename[extension + 1:].lower()
            if extension == 'jpg':
                extension = 'jpeg'
            assert extension in self.extensions
            if extension == 'bmp':
                width, height = self.__bmp(value.file)
            elif extension == 'gif':
                width, height = self.__gif(value.file)
            elif extension == 'jpeg':
                width, height = self.__jpeg(value.file)
            elif extension == 'png':
                width, height = self.__png(value.file)
            else:
                width = -1
                height = -1
            assert self.minsize[0] <= width <= self.maxsize[0] \
                and self.minsize[1] <= height <= self.maxsize[1]
            value.file.seek(0)
            return value, None
        except:
            return value, translate(self.message)

    def __bmp(self, stream):
        if stream.read(2) == 'BM':
            stream.read(16)
            return struct.unpack("<LL", stream.read(8))
        return -1, -1

    def __gif(self, stream):
        if stream.read(6) in ('GIF87a', 'GIF89a'):
            stream = stream.read(5)
            if len(stream) == 5:
                return tuple(struct.unpack("<HHB", stream)[:-1])
        return -1, -1

    def __jpeg(self, stream):
        if stream.read(2) == '\xFF\xD8':
            while True:
                (marker, code, length) = struct.unpack("!BBH", stream.read(4))
                if marker != 0xFF:
                    break
                elif code >= 0xC0 and code <= 0xC3:
                    return tuple(reversed(
                        struct.unpack("!xHH", stream.read(5))))
                else:
                    stream.read(length - 2)
        return -1, -1

    def __png(self, stream):
        if stream.read(8) == '\211PNG\r\n\032\n':
            stream.read(4)
            if stream.read(4) == "IHDR":
                return struct.unpack("!LL", stream.read(8))
        return -1, -1


class _isGenericUrl(Validator):
    """
    Rejects a URL string if any of the following is true:
       * The string is empty or None
       * The string uses characters that are not allowed in a URL
       * The URL scheme specified (if one is specified) is not valid

    Based on RFC 2396: http://www.faqs.org/rfcs/rfc2396.html
    @author: Jonathan Benn
    """
    message = "Invalid URL"
    all_url_schemes = [None] + official_url_schemes + unofficial_url_schemes

    def __init__(self, schemes=None, prepend_scheme=None, message=None):
        super().__init__(message=message)
        self.allowed_schemes = schemes or self.all_url_schemes
        self.prepend_scheme = prepend_scheme
        if self.prepend_scheme not in self.allowed_schemes:
            raise SyntaxError(
                "prepend_scheme='{}' is not in allowed_schemes={}".format(
                    self.prepend_scheme, self.allowed_schemes
                )
            )

    GENERIC_URL = re.compile(
        r"%[^0-9A-Fa-f]{2}|%[^0-9A-Fa-f][0-9A-Fa-f]|%[0-9A-Fa-f][^0-9A-Fa-f]|"
        r"%$|%[0-9A-Fa-f]$|%[^0-9A-Fa-f]$"
    )
    GENERIC_URL_VALID = re.compile(r"[A-Za-z0-9;/?:@&=+$,\-_\.!~*'\(\)%#]+$")

    def __call__(self, value):
        try:
            # if the URL does not misuse the '%' character
            if not self.GENERIC_URL.search(value):
                # if the URL is only composed of valid characters
                if self.GENERIC_URL_VALID.match(value):
                    # Then split up the URL into its components and check on
                    # the scheme
                    scheme = url_split_regex.match(value).group(2)
                    # Clean up the scheme before we check it
                    if scheme is not None:
                        scheme = url_unquote(scheme).lower()
                    # If the scheme really exists
                    if scheme in self.allowed_schemes:
                        # Then the URL is valid
                        return value, None
                    else:
                        # else, for the possible case of abbreviated URLs with
                        # ports, check to see if adding a valid scheme fixes
                        # the problem (but only do this if it doesn't have
                        # one already!)
                        if value.find('://') < 0 and \
                                None in self.allowed_schemes:
                            schemeToUse = self.prepend_scheme or 'http'
                            prependTest = self.__call__(
                                schemeToUse + '://' + value)
                            # if the prepend test succeeded
                            if prependTest[1] is None:
                                # if prepending in the output is enabled
                                if self.prepend_scheme:
                                    return prependTest
                                else:
                                    # else return the original,
                                    #  non-prepended value
                                    return value, None
        except Exception:
            pass
        # else the URL is not valid
        return value, translate(self.message)


class _isHTTPUrl(Validator):
    """
    Rejects a URL string if any of the following is true:
       * The string is empty or None
       * The string uses characters that are not allowed in a URL
       * The string breaks any of the HTTP syntactic rules
       * The URL scheme specified (if one is specified) is not 'http' or
         'https'
       * The top-level domain (if a host name is specified) does not exist

    Based on RFC 2616: http://www.faqs.org/rfcs/rfc2616.html
    @author: Jonathan Benn
    """

    message = "Invalid URL"
    http_schemes = [None, 'http', 'https']
    GENERIC_VALID_IP = re.compile(
        r"([\w.!~*'|;:&=+$,-]+@)?\d+\.\d+\.\d+\.\d+(:\d*)*$"
    )
    GENERIC_VALID_DOMAIN = re.compile(
        r"([\w.!~*'|;:&=+$,-]+@)?(([A-Za-z0-9]+[A-Za-z0-9\-]*[A-Za-z0-9]+\.)"
        r"*([A-Za-z0-9]+\.)*)*([A-Za-z]+[A-Za-z0-9\-]*[A-Za-z0-9]+)\.?(:\d*)*$"
    )

    def __init__(self, schemes=None, prepend_scheme='http', tlds=None, message=None):
        super().__init__(message=message)
        self.allowed_schemes = schemes or self.http_schemes
        self.allowed_tlds = tlds or official_top_level_domains
        self.prepend_scheme = prepend_scheme
        for i in self.allowed_schemes:
            if i not in self.http_schemes:
                raise SyntaxError(
                    "allowed_scheme value '{}' is not in {}".format(
                        i, self.http_schemes
                    )
                )
        if self.prepend_scheme not in self.allowed_schemes:
            raise SyntaxError(
                "prepend_scheme='{}' is not in allowed_schemes={}".format(
                    self.prepend_scheme, self.allowed_schemes
                )
            )

    def __call__(self, value):
        try:
            # if the URL passes generic validation
            x = _isGenericUrl(
                schemes=self.allowed_schemes,
                prepend_scheme=self.prepend_scheme,
                message=self.message
            )
            if x(value)[1] is None:
                componentsMatch = url_split_regex.match(value)
                authority = componentsMatch.group(4)
                # if there is an authority component
                if authority:
                    # if authority is a valid IP address
                    if self.GENERIC_VALID_IP.match(authority):
                        # Then this HTTP URL is valid
                        return value, None
                    else:
                        # else if authority is a valid domain name
                        domainMatch = self.GENERIC_VALID_DOMAIN.match(
                            authority)
                        if domainMatch:
                            # if the top-level domain really exists
                            if domainMatch.group(5).lower()\
                                    in self.allowed_tlds:
                                # Then this HTTP URL is valid
                                return value, None
                else:
                    # else this is a relative/abbreviated URL, which will parse
                    # into the URL's path component
                    path = componentsMatch.group(5)
                    # relative case: if this is a valid path (if it starts with
                    # a slash)
                    if path.startswith('/'):
                        # Then this HTTP URL is valid
                        return value, None
                    else:
                        # abbreviated case: if we haven't already, prepend a
                        # scheme and see if it fixes the problem
                        if value.find('://') < 0:
                            schemeToUse = self.prepend_scheme or 'http'
                            prependTest = self(schemeToUse + '://' + value)
                            # if the prepend test succeeded
                            if prependTest[1] is None:
                                # if prepending in the output is enabled
                                if self.prepend_scheme:
                                    return prependTest
                                else:
                                    # else return the original, non-prepended
                                    # value
                                    return value, None
        except:
            pass
        # else the HTTP URL is not valid
        return value, translate(self.message)


class isUrl(Validator):
    #: use `_isGenericUrl` and `_isHTTPUrl` depending on `mode` parameter
    message = "Invalid URL"

    def __init__(
        self, mode='http', schemes=None, prepend_scheme='http', tlds=None, message=None
    ):
        super().__init__(message=message)
        self.mode = mode.lower()
        if self.mode not in ['generic', 'http']:
            raise SyntaxError("invalid mode '{}' in isUrl".format(self.mode))
        self.allowed_tlds = tlds
        self.allowed_schemes = schemes
        if self.allowed_schemes:
            if prepend_scheme not in self.allowed_schemes:
                raise SyntaxError(
                    "prepend_scheme='{}' is not in allowed_schemes={}".format(
                        prepend_scheme, self.allowed_schemes
                    )
                )
        # if allowed_schemes is None, then we will defer testing
        # prepend_scheme's validity to a sub-method
        self.prepend_scheme = prepend_scheme

    def __call__(self, value):
        if self.mode == 'generic':
            subValidator = _isGenericUrl(
                schemes=self.allowed_schemes,
                prepend_scheme=self.prepend_scheme,
                message=self.message
            )
        elif self.mode == 'http':
            subValidator = _isHTTPUrl(
                schemes=self.allowed_schemes,
                prepend_scheme=self.prepend_scheme,
                tlds=self.allowed_tlds,
                message=self.message
            )
        else:
            raise SyntaxError("invalid mode '{}' in isUrl".format(self.mode))

        return subValidator(value)


class isIPv4(Validator):
    """
    Checks if field's value is an IP version 4 address in decimal form. Can
    be set to force addresses from certain range.

    IPv4 regex taken from: http://regexlib.com/REDetails.aspx?regexp_id=1411

    Args:
        min: lowest allowed address; accepts:
            - str, eg. 192.168.0.1
            - list or tuple of octets, eg. [192, 168, 0, 1]
        max: highest allowed address; same as above
        invert: True to allow addresses only from outside of given range; note
            that range boundaries are not matched this way
        localhost: localhost address treatment:
            - None (default): indifferent
            - True (enforce): query address must match localhost address
            - False (forbid): query address must not match localhost address
        private: same as above, except that query address is checked against
            two address ranges: 172.16.0.0 - 172.31.255.255 and
            192.168.0.0 - 192.168.255.255
        automatic: same as above, except that query address is checked
            against one address range: 169.254.0.0 - 169.254.255.255

    Minip and maxip may also be lists or tuples of addresses in all above
    forms (str, int, list / tuple), allowing setup of multiple address ranges::
        minip = (minip1, minip2, ... minipN)
                   |       |           |
                   |       |           |
        maxip = (maxip1, maxip2, ... maxipN)

    Longer iterable will be truncated to match length of shorter one.
    """

    message = "Invalid IPv4 address"
    regex = re.compile(
        r"^(([1-9]?\d|1\d\d|2[0-4]\d|25[0-5])\.){3}([1-9]?\d|1\d\d|2[0-4]\d|25[0-5])$"
    )
    numbers = (16777216, 65536, 256, 1)
    localhost = 2130706433
    private = ((2886729728, 2886795263), (3232235520, 3232301055))
    automatic = (2851995648, 2852061183)

    def __init__(
        self,
        min='0.0.0.0',
        max='255.255.255.255',
        invert=False,
        localhost=None,
        private=None,
        auto=None,
        message=None
    ):
        super().__init__(message=message)
        for n, value in enumerate((min, max)):
            temp = []
            if isinstance(value, str):
                temp.append(value.split('.'))
            elif isinstance(value, (list, tuple)):
                if len(value) == len(
                    list(filter(lambda item: isinstance(item, int), value))
                ) == 4:
                    temp.append(value)
                else:
                    for item in value:
                        if isinstance(item, str):
                            temp.append(item.split('.'))
                        elif isinstance(item, (list, tuple)):
                            temp.append(item)
            numbers = []
            for item in temp:
                number = 0
                for i, j in zip(self.numbers, item):
                    number += i * int(j)
                numbers.append(number)
            if n == 0:
                self.minip = numbers
            else:
                self.maxip = numbers
        self.invert = invert
        self.is_localhost = localhost
        self.is_private = private
        self.is_automatic = auto

    def __call__(self, value):
        if self.regex.match(value):
            number = 0
            for i, j in zip(self.numbers, value.split('.')):
                number += i * int(j)
            ok = False
            for bottom, top in zip(self.minip, self.maxip):
                if self.invert != (bottom <= number <= top):
                    ok = True
            if not (
                self.is_localhost is None or self.is_localhost == (
                    number == self.localhost
                )
            ):
                ok = False
            if not (
                self.is_private is None or self.is_private == (
                    sum([
                        el[0] <= number <= el[1]
                        for el in self.private
                    ]) > 0
                )
            ):
                ok = False
            if not (
                self.is_automatic is None or self.is_automatic == (
                    self.automatic[0] <= number <= self.automatic[1]
                )
            ):
                ok = False
            if ok:
                return value, None
        return value, translate(self.message)


class isIPv6(Validator):
    """
    Checks if field's value is an IP version 6 address.
    Use the ipaddress library.

    Args:
        private: None (default): indifferent
                 True (enforce): address must be in fc00::/7 range
                 False (forbid): address must NOT be in fc00::/7 range
        link_local: Same as above but uses fe80::/10 range
        reserved: Same as above but uses IETF reserved range
        mulicast: Same as above but uses ff00::/8 range
        routable: Similar to above but enforces not private, link_local,
                   reserved or multicast
        to4: Same as above but uses 2002::/16 range
        teredo: Same as above but uses 2001::/32 range
        subnets: value must be a member of at least one from list of subnets
    """

    message = "Invalid IPv6 address"

    def __init__(
        self,
        private=None,
        link_local=None,
        reserved=None,
        multicast=None,
        routable=None,
        to4=None,
        teredo=None,
        subnets=None,
        message=None
    ):
        super().__init__(message=message)
        self.is_private = private
        self.is_link_local = link_local
        self.is_reserved = reserved
        self.is_multicast = multicast
        self.is_routable = routable
        self.is_6to4 = to4
        self.is_teredo = teredo
        self.subnets = subnets

        if ipaddress is None:
            raise RuntimeError(
                "You need 'ipaddress' python module to use isIPv6 validator."
            )

    def __call__(self, value):
        try:
            ip = ipaddress.IPv6Address(value)
            ok = True
        except ipaddress.AddressValueError:
            return (value, translate(self.message))

        if self.subnets:
            # iterate through self.subnets to see if value is a member
            ok = False
            if isinstance(self.subnets, str):
                self.subnets = [self.subnets]
            for network in self.subnets:
                try:
                    ipnet = ipaddress.IPv6Network(network)
                except (ipaddress.NetmaskValueError,
                        ipaddress.AddressValueError):
                    return value, translate('invalid subnet provided')
                if ip in ipnet:
                    ok = True

        if self.is_routable:
            self.is_private = False
            self.is_link_local = False
            self.is_reserved = False
            self.is_multicast = False

        if not (
            self.is_private is None or self.is_private == ip.is_private
        ):
            ok = False
        if not (
            self.is_link_local is None or
            self.is_link_local == ip.is_link_local
        ):
            ok = False
        if not (
            self.is_reserved is None or self.is_reserved == ip.is_reserved
        ):
            ok = False
        if not (
            self.is_multicast is None or self.is_multicast == ip.is_multicast
        ):
            ok = False
        if not (self.is_6to4 is None or self.is_6to4 == ip.is_6to4):
            ok = False
        if not (self.is_teredo is None or self.is_teredo == ip.is_teredo):
            ok = False

        if ok:
            return value, None

        return value, translate(self.message)


class isIP(Validator):
    """
    Checks if field's value is an IP Address (v4 or v6). Can be set to force
    addresses from within a specific range. Checks are done with the correct
    isIPv4 and isIPv6 validators.

    Uses ipaddress library.
    """

    message = "Invalid IP address"

    def __init__(
        self,
        min='0.0.0.0',
        max='255.255.255.255',
        invert=False,
        localhost=None,
        private=None,
        auto=None,
        ipv4=None,
        link_local=None,
        reserved=None,
        multicast=None,
        routable=None,
        to4=None,
        teredo=None,
        subnets=None,
        ipv6=None,
        message=None
    ):
        super().__init__(message=message)
        self.minip = min,
        self.maxip = max,
        self.invert = invert
        self.is_localhost = localhost
        self.is_private = private
        self.is_automatic = auto
        self.is_ipv4 = ipv4
        self.is_private = private
        self.is_link_local = link_local
        self.is_reserved = reserved
        self.is_multicast = multicast
        self.is_routable = routable
        self.is_6to4 = to4
        self.is_teredo = teredo
        self.subnets = subnets
        self.is_ipv6 = ipv6

        if ipaddress is None:
            raise RuntimeError(
                "You need 'ipaddress' python module to use isIP validator."
            )

    def __call__(self, value):
        try:
            ip = ipaddress.ip_address(value)
        except ValueError:
            return value, translate(self.message)

        if self.is_ipv4 and isinstance(ip, ipaddress.IPv6Address):
            rv = (value, translate(self.message))
        elif self.is_ipv6 and isinstance(ip, ipaddress.IPv4Address):
            rv = (value, translate(self.message))
        elif self.is_ipv4 or isinstance(ip, ipaddress.IPv4Address):
            rv = isIPv4(
                min=self.minip,
                max=self.maxip,
                invert=self.invert,
                localhost=self.is_localhost,
                private=self.is_private,
                auto=self.is_automatic,
                message=self.message
            )(value)
        elif self.is_ipv6 or isinstance(ip, ipaddress.IPv6Address):
            rv = isIPv6(
                private=self.is_private,
                link_local=self.is_link_local,
                reserved=self.is_reserved,
                multicast=self.is_multicast,
                routable=self.is_routable,
                to4=self.is_6to4,
                teredo=self.is_teredo,
                subnets=self.subnets,
                message=self.message
            )(value)
        else:
            rv = (value, translate(self.message))
        return rv



# -*- coding: utf-8 -*-
"""
    tests.mailer
    ------------

    Test Emmett mailer
"""

import base64
import email
import pytest
import re
import time

from email.header import Header
from emmett import App
from emmett.tools.mailer import Mailer, sanitize_address


@pytest.fixture(scope='module')
def app():
    rv = App(__name__)
    rv.config.mailer.sender = 'support@example.com'
    return rv


@pytest.fixture(scope='module')
def mailer(app):
    rv = Mailer(app)
    return rv


def test_message_init(mailer):
    msg = mailer.mail(subject="subject", recipients="to@example.com")
    assert msg.sender == 'support@example.com'
    assert msg.recipients == ["to@example.com"]


def test_recipients(mailer):
    msg = mailer.mail(subject="subject")
    assert msg.recipients == []
    msg.add_recipient("to@example.com")
    assert msg.recipients == ["to@example.com"]


def test_all_recipients(mailer):
    msg = mailer.mail(
        subject="subject", recipients=["somebody@here.com"],
        cc=["cc@example.com"], bcc=["bcc@example.com"])
    assert len(msg.all_recipients) == 3
    msg.add_recipient("cc@example.com")
    assert len(msg.all_recipients) == 3


def test_sender(mailer):
    msg = mailer.mail(
        subject="subject",
        sender=u'ÄÜÖ → ✓ <from@example.com>>',
        recipients=["to@example.com"])
    assert 'From: =?utf-8?b?w4TDnMOWIOKGkiDinJM=?= <from@example.com>' in \
        str(msg)


def test_sender_as_tuple(mailer):
    msg = mailer.mail(
        subject="testing", sender=("tester", "tester@example.com"))
    assert msg.sender == "tester <tester@example.com>"
    msg = mailer.mail(
        subject="subject",
        sender=(u"ÄÜÖ → ✓", 'from@example.com>'),
        recipients=["to@example.com"])
    assert 'From: =?utf-8?b?w4TDnMOWIOKGkiDinJM=?= <from@example.com>' in \
        str(msg)


def test_reply_to(mailer):
    msg = mailer.mail(
        subject="testing",
        recipients=["to@example.com"],
        sender="spammer <spammer@example.com>",
        reply_to="somebody <somebody@example.com>",
        body="testing")
    h = Header(
        "Reply-To: %s" % sanitize_address('somebody <somebody@example.com>'))
    assert h.encode() in str(msg)


def test_missing_sender(mailer):
    msg = mailer.mail(
        subject="testing", recipients=["to@example.com"], body="testing")
    msg.sender = None
    with pytest.raises(AssertionError):
        msg.send()


def test_missing_recipients(mailer):
    msg = mailer.mail(subject="testing", recipients=[], body="testing")
    with pytest.raises(AssertionError):
        msg.send()


def test_bcc(mailer):
    msg = mailer.mail(
        sender="from@example.com",
        subject="testing",
        recipients=["to@example.com"],
        body="testing",
        bcc=["tosomeoneelse@example.com"])
    assert "tosomeoneelse@example.com" not in str(msg)


def test_cc(mailer):
    msg = mailer.mail(
        sender="from@example.com",
        subject="testing",
        recipients=["to@example.com"],
        body="testing",
        cc=["tosomeoneelse@example.com"])
    assert "Cc: tosomeoneelse@example.com" in str(msg)


def test_attach(mailer):
    msg = mailer.mail(
        subject="testing", recipients=["to@example.com"], body="testing")
    msg.attach(data=b"this is a test", content_type="text/plain")
    a = msg.attachments[0]
    assert a.filename is None
    assert a.disposition == 'attachment'
    assert a.content_type == 'text/plain'
    assert a.data == b"this is a test"


def test_bad_subject(mailer):
    msg = mailer.mail(
        subject="testing\r\n",
        sender="from@example.com",
        body="testing",
        recipients=["to@example.com"])
    with pytest.raises(RuntimeError):
        msg.send()


def test_subject(mailer):
    msg = mailer.mail(
        subject=u"sübject",
        sender='from@example.com',
        recipients=["to@example.com"])
    assert '=?utf-8?q?s=C3=BCbject?=' in str(msg)


def test_empty_subject(mailer):
    msg = mailer.mail(sender="from@example.com", recipients=["foo@bar.com"])
    msg.body = "normal ascii text"
    assert 'Subject:' not in str(msg)


def test_multiline_subject(mailer):
    msg = mailer.mail(
        subject="testing\r\n testing\r\n testing \r\n \ttesting",
        sender="from@example.com",
        body="testing",
        recipients=["to@example.com"])
    msg_as_string = str(msg)
    assert "From: from@example.com" in msg_as_string
    assert "testing\r\n testing\r\n testing \r\n \ttesting" in msg_as_string
    msg = mailer.mail(
        subject="testing\r\n testing\r\n ",
        sender="from@example.com",
        body="testing",
        recipients=["to@example.com"])
    with pytest.raises(RuntimeError):
        msg.send()
    msg = mailer.mail(
        subject="testing\r\n testing\r\n\t",
        sender="from@example.com",
        body="testing",
        recipients=["to@example.com"])
    with pytest.raises(RuntimeError):
        msg.send()
    msg = mailer.mail(
        subject="testing\r\n testing\r\n\n",
        sender="from@example.com",
        body="testing",
        recipients=["to@example.com"])
    with pytest.raises(RuntimeError):
        msg.send()


def test_bad_sender(mailer):
    msg = mailer.mail(
        subject="testing",
        sender="from@example.com\r\n",
        recipients=["to@example.com"],
        body="testing")
    assert 'From: from@example.com' in str(msg)


def test_bad_reply_to(mailer):
    msg = mailer.mail(
        subject="testing",
        sender="from@example.com",
        reply_to="evil@example.com\r",
        recipients=["to@example.com"],
        body="testing")
    msg_as_string = str(msg)
    assert 'From: from@example.com' in msg_as_string
    assert 'To: to@example.com' in msg_as_string
    assert 'Reply-To: evil@example.com' in msg_as_string


def test_bad_recipient(mailer):
    msg = mailer.mail(
        subject="testing",
        sender="from@example.com",
        recipients=["to@example.com", "to\r\n@example.com"],
        body="testing")
    assert 'To: to@example.com' in str(msg)


def test_address_sanitize(mailer):
    msg = mailer.mail(
        subject="testing",
        sender="sender\r\n@example.com",
        reply_to="reply_to\r\n@example.com",
        recipients=["recipient\r\n@example.com"])
    msg_as_string = str(msg)
    assert 'sender@example.com' in msg_as_string
    assert 'reply_to@example.com' in msg_as_string
    assert 'recipient@example.com' in msg_as_string


def test_plain_message(mailer):
    plain_text = "Hello Joe,\nHow are you?"
    msg = mailer.mail(
        sender="from@example.com",
        subject="subject",
        recipients=["to@example.com"],
        body=plain_text)
    assert plain_text == msg.body
    assert 'Content-Type: text/plain' in str(msg)


def test_plain_message_with_attachments(mailer):
    msg = mailer.mail(
        sender="from@example.com",
        subject="subject",
        recipients=["to@example.com"],
        body="hello")
    msg.attach(data=b"this is a test", content_type="text/plain")
    assert 'Content-Type: multipart/mixed' in str(msg)


def test_plain_message_with_unicode_attachments(mailer):
    msg = mailer.mail(
        subject="subject", recipients=["to@example.com"], body="hello")
    msg.attach(
        data=b"this is a test",
        content_type="text/plain",
        filename=u'ünicöde ←→ ✓.txt')
    parsed = email.message_from_string(str(msg))
    assert (
        re.sub(r'\s+', ' ', parsed.get_payload()[1].get('Content-Disposition'))
        in [
            'attachment; filename*="UTF8\'\''
            '%C3%BCnic%C3%B6de%20%E2%86%90%E2%86%92%20%E2%9C%93.txt"',
            'attachment; filename*=UTF8\'\''
            '%C3%BCnic%C3%B6de%20%E2%86%90%E2%86%92%20%E2%9C%93.txt'
        ])


def test_html_message(mailer):
    html_text = "<p>Hello World</p>"
    msg = mailer.mail(
        sender="from@example.com",
        subject="subject",
        recipients=["to@example.com"],
        html=html_text)
    assert html_text == msg.html
    assert 'Content-Type: multipart/alternative' in str(msg)


def test_json_message(mailer):
    json_text = '{"msg": "Hello World!}'
    msg = mailer.mail(
        sender="from@example.com",
        subject="subject",
        recipients=["to@example.com"],
        alts={'json': json_text})
    assert json_text == msg.alts['json']
    assert 'Content-Type: multipart/alternative' in str(msg)


def test_html_message_with_attachments(mailer):
    html_text = "<p>Hello World</p>"
    plain_text = 'Hello World'
    msg = mailer.mail(
        sender="from@example.com",
        subject="subject",
        recipients=["to@example.com"],
        body=plain_text,
        html=html_text)
    msg.attach(data=b"this is a test", content_type="text/plain")
    assert html_text == msg.html
    assert 'Content-Type: multipart/alternative' in str(msg)
    parsed = email.message_from_string(str(msg))
    assert len(parsed.get_payload()) == 2
    body, attachment = parsed.get_payload()
    assert len(body.get_payload()) == 2
    plain, html = body.get_payload()
    assert plain.get_payload() == plain_text
    assert html.get_payload() == html_text
    assert base64.b64decode(attachment.get_payload()) == b'this is a test'


def test_date(mailer):
    before = time.time()
    msg = mailer.mail(
        sender="from@example.com",
        subject="subject",
        recipients=["to@example.com"],
        body="hello",
        date=time.time())
    after = time.time()
    assert before <= msg.date <= after
    fmt_date = email.utils.formatdate(msg.date, localtime=True)
    assert 'Date: {}'.format(fmt_date) in str(msg)


def test_msgid(mailer):
    msg = mailer.mail(
        sender="from@example.com",
        subject="subject",
        recipients=["to@example.com"],
        body="hello")
    r = re.compile(r"<\S+@\S+>").match(msg.msgId)
    assert r is not None
    assert 'Message-ID: {}'.format(msg.msgId) in str(msg)


def test_unicode_addresses(mailer):
    msg = mailer.mail(
        subject="subject",
        sender=u'ÄÜÖ → ✓ <from@example.com>',
        recipients=[u"Ä <t1@example.com>", u"Ü <t2@example.com>"],
        cc=[u"Ö <cc@example.com>"])
    msg_as_string = str(msg)
    a1 = sanitize_address(u"Ä <t1@example.com>")
    a2 = sanitize_address(u"Ü <t2@example.com>")
    h1_a = Header("To: %s, %s" % (a1, a2))
    h1_b = Header("To: %s, %s" % (a2, a1))
    h2 = Header("From: %s" % sanitize_address(u"ÄÜÖ → ✓ <from@example.com>"))
    h3 = Header("Cc: %s" % sanitize_address(u"Ö <cc@example.com>"))
    try:
        assert h1_a.encode() in msg_as_string
    except AssertionError:
        assert h1_b.encode() in msg_as_string
    assert h2.encode() in msg_as_string
    assert h3.encode() in msg_as_string


def test_extra_headers(mailer):
    msg = mailer.mail(
        sender="from@example.com",
        subject="subject",
        recipients=["to@example.com"],
        body="hello",
        extra_headers={'X-Extra-Header': 'Yes'})
    assert 'X-Extra-Header: Yes' in str(msg)


def test_send(mailer):
    with mailer.store_mails() as outbox:
        msg = mailer.mail(
            subject="testing", recipients=["tester@example.com"], body="test")
        msg.send()
        assert msg.date is not None
        assert len(outbox) == 1
        sent_msg = outbox[0]
        assert sent_msg.sender == 'support@example.com'


def test_send_mail(mailer):
    with mailer.store_mails() as outbox:
        mailer.send_mail(
            subject="testing", recipients=["tester@example.com"], body="test")
        assert len(outbox) == 1
        sent_msg = outbox[0]
        assert sent_msg.subject == 'testing'
        assert sent_msg.recipients == ["tester@example.com"]
        assert sent_msg.body == 'test'
        assert sent_msg.sender == 'support@example.com'



# -*- coding: utf-8 -*-
"""
    tests.orm_transactions
    ----------------------

    Test pyDAL transactions implementation over Emmett.
"""

import pytest

from emmett import App, sdict
from emmett.orm import Database, Field, Model


class Register(Model):
    value = Field.int()


@pytest.fixture(scope='module')
def db():
    app = App(__name__)
    db = Database(
        app, config=sdict(
            uri='sqlite:memory', auto_migrate=True, auto_connect=True))
    db.define_models(Register)
    return db


@pytest.fixture(scope='function')
def cleanup(request, db):
    def teardown():
        Register.all().delete()
        db.commit()

    request.addfinalizer(teardown)


def _save(*vals):
    for val in vals:
        Register.create(value=val)


def _values_in_register(*vals):
    db_vals = Register.all().select(orderby=Register.value).column('value')
    return db_vals == list(vals)


def test_transactions(db, cleanup):
    adapter = db._adapter

    assert adapter.in_transaction()

    _save(1)
    db.commit()
    assert _values_in_register(1)

    _save(2)
    db.rollback()
    _save(3)
    with db.atomic():
        _save(4)
        with db.atomic() as sp2:
            _save(5)
            sp2.rollback()
        with db.atomic():
            _save(6)
            with db.atomic() as sp4:
                _save(7)
                with db.atomic():
                    _save(8)
                assert _values_in_register(1, 3, 4, 6, 7, 8)
                sp4.rollback()
            assert _values_in_register(1, 3, 4, 6)
    db.commit()
    assert _values_in_register(1, 3, 4, 6)


def _commit_rollback(db):
    _save(1)
    db.commit()
    _save(2)
    db.rollback()
    assert _values_in_register(1)

    _save(3)
    db.rollback()
    _save(4)
    db.commit()
    assert _values_in_register(1, 4)


def test_commit_rollback(db, cleanup):
    _commit_rollback(db)


def test_commit_rollback_nested(db, cleanup):
    with db.atomic():
        _commit_rollback(db)
    assert _values_in_register(1, 4)


def test_nested_transaction_obj(db, cleanup):
    assert _values_in_register()

    _save(1)
    with db.transaction() as txn:
        _save(2)
        txn.rollback()
        assert _values_in_register()
    _save(3)
    db.commit()
    assert _values_in_register(3)


def test_savepoint_commit(db, cleanup):
    _save(1)
    db.rollback()

    _save(2)
    db.commit()

    with db.atomic() as sp:
        _save(3)
        sp.rollback()
        _save(4)
        sp.commit()

    assert _values_in_register(2, 4)


def text_atomic_exception(db, cleanup):
    def will_fail():
        with db.atomic():
            _save(1)
            _save(None)

    with pytest.raises(Exception):
        will_fail()
    assert _values_in_register()

    def user_error():
        with db.atomic():
            _save(2)
            raise ValueError

    with pytest.raises(ValueError):
        user_error()
    assert _values_in_register()
